import os
import argparse
import torch
from sklearn.metrics import roc_auc_score
import numpy as np
from maldi_nn.models import AMRModel
from maldi_nn.utils.data import DRIAMSAMRDataModule
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch import Trainer
from importlib.resources import files
import json
import ast
import h5torch


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


# eval phase:
def micro_roc_auc(trues, preds, locs, drug_names):
    return roc_auc_score(trues, preds)


def macro_roc_auc(trues, preds, locs, drug_names):
    t = []
    for c in np.unique(drug_names):
        trues_sub = trues[drug_names == c]
        if len(np.unique(trues_sub)) > 1:
            preds_sub = preds[drug_names == c]
            t.append(roc_auc_score(trues_sub, preds_sub))
    return np.mean(t)


def instance_roc_auc(trues, preds, locs, drug_names):
    r, c = np.unique(locs, return_counts=True)
    pos = 0
    total = 0
    for r_ in r[c > 1]:
        trues_sub = trues[locs == r_]
        if len(np.unique(trues[locs == r_])) > 1:
            preds_sub = preds[locs == r_]
            p = preds_sub[trues_sub == 1][:, None] >= preds_sub[trues_sub == 0]
            pos += p.sum()
            total += p.size
    return pos / total


def prec_at_1_neg(trues, preds, locs, drug_names):
    r, c = np.unique(locs, return_counts=True)
    pos = 0
    total = 0
    for r_ in r[c > 1]:
        trues_sub = trues[locs == r_]
        if len(np.unique(trues[locs == r_])) > 1:
            preds_sub = preds[locs == r_]
            pos += trues_sub[np.argmin(preds_sub)]
            total += 1
    return (total - pos) / total


size_to_layer_dims = {
    "S": [256, 128],
    "M": [512, 256, 128],
    "L": [1024, 512, 256, 128],
    "XL": [2048, 1024, 512, 256, 128],
    "Linear": [],
}

drug_encoder_args = {
    "onehot": {},
    "kernel": {},
    "ecfp": {"bits": 512, "diameter": 4},
    "cnn": {"alphabet": "deepsmiles"},
    "trf": {"alphabet": "deepsmiles"},
    "gru": {"alphabet": "deepsmiles"},
    "img": {"size": (128, 128), "normalize": True, "inverse": True},
}


def main():
    parser = argparse.ArgumentParser(
        description="Training script for dual-branch AMR recommender fine-tuning on other DRIAMS hospitals.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument(
        "logs_path", type=str, metavar="logs_path", help="path to logs."
    )
    parser.add_argument(
        "drug_embedder",
        type=str,
        metavar="drug_embedder",
        choices=["ecfp", "onehot", "gru", "cnn", "trf", "img", "kernel"],
        help="Which drug embedder to use, choices: {%(choices)s} Ignored if a ckpt_path is given.",
    )
    parser.add_argument(
        "spectrum_embedder",
        type=str,
        metavar="spectrum_embedder",
        choices=["S", "M", "L", "XL", "Linear"],
        help="Which size spectrum embedder to use, choices: {%(choices)s} Ignored if a ckpt_path is given",
    )
    parser.add_argument(
        "hospital",
        type=str,
        metavar="hospital",
        choices=["B", "C", "D"],
        help="Which DRIAMS hospital to fine-tune on, choices: {%(choices)s}",
    )

    parser.add_argument(
        "percent",
        type=float,
        metavar="percent",
        help="Percentage of training data to use (100percent means 1000 samples)",
    )

    parser.add_argument(
        "--ckpt_path",
        type=str,
        default="None",
        help="Checkpoint from which to start training",
    )

    parser.add_argument(
        "--lr",
        type=float,
        default=5e-4,
        help="Learning rate, Ignored if a ckpt_path is given.",
    )
    parser.add_argument(
        "--logging_file",
        type=str,
        default="res.txt",
        help="Which file to write final performances to.",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=4,
        help="Number of workers in dataloader. Reduce to alleviate CPU.",
    )

    parser.add_argument(
        "--devices",
        type=ast.literal_eval,
        default=1,
        help="devices to use. Input an integer to specify a number of gpus or a list e.g. [1] or [0,1,3] to specify which gpus.",
    )

    args = parser.parse_args()

    dm = DRIAMSAMRDataModule(
        args.path,
        drug_encoder=args.drug_embedder,
        drug_encoder_args=drug_encoder_args[args.drug_embedder],
        batch_size=128,
        n_workers=args.num_workers,
        preprocessor=None,
        min_spectrum_len=None,
        in_memory=False,
    )
    dm.setup(None)

    split_path = files("maldi_nn.utils").joinpath("hospital_split.json")
    hosp_split = json.load(open(split_path))
    train_locs = np.array(hosp_split[args.hospital + "_train"]).astype(bytes)[
        : int(1000 * args.percent)
    ]
    val_locs = np.array(hosp_split[args.hospital + "_val"]).astype(bytes)
    test_locs = np.array(hosp_split[args.hospital + "_test"]).astype(bytes)

    f = h5torch.File(args.path)
    train_rows = np.where(np.isin(f["0/loc"][:], train_locs))[0]
    train_indices = np.isin(f["central/indices"][:][0], train_rows)
    val_rows = np.where(np.isin(f["0/loc"][:], val_locs))[0]
    val_indices = np.isin(f["central/indices"][:][0], val_rows)
    test_rows = np.where(np.isin(f["0/loc"][:], test_locs))[0]
    test_indices = np.isin(f["central/indices"][:][0], test_rows)

    if dm.min_spectrum_len is not None:
        lens_all = np.array([len(tt) for tt in f["0/intensity"][:]])

        lens = lens_all[f["central"]["indices"][0][train_indices]]
        train_indices = train_indices[lens > dm.min_spectrum_len]

        lens = lens_all[f["central"]["indices"][0][val_indices]]
        val_indices = val_indices[lens > dm.min_spectrum_len]

        lens = lens_all[f["central"]["indices"][0][test_indices]]
        test_indices = test_indices[lens > dm.min_spectrum_len]

    f = f.to_dict()

    dm.train = h5torch.Dataset(
        f,
        sampling="coo",
        subset=train_indices,
        sample_processor=dm.processor,
    )

    dm.val = h5torch.Dataset(
        f,
        sampling="coo",
        subset=val_indices,
        sample_processor=dm.processor,
    )

    dm.test = h5torch.Dataset(
        f,
        sampling="coo",
        subset=test_indices,
        sample_processor=dm.processor,
    )

    print(len(dm.train), len(dm.val), len(dm.test))

    drug_embedder_kwargs = {
        "onehot": {
            "num_drugs": len(dm.drug_encoder.vocab),
            "dim": 64,
        },
        "kernel": {
            "n_inputs": len(dm.drug_encoder.embeds),
            "n_outputs": 64,
            "layer_dims": [],
            "dropout": 0.2,
        },
        "ecfp": {"n_inputs": 512, "n_outputs": 64, "layer_dims": [], "dropout": 0.2},
        "cnn": {
            "vocab_size": len(dm.drug_encoder.alphabet_dict),
            "dim": 64,
            "depth": 4,
            "kernel_size": 5,
        },
        "trf": {
            "vocab_size": len(dm.drug_encoder.alphabet_dict),
            "dim": 64,
            "depth": 4,
            "n_head": 8,
        },
        "gru": {"vocab_size": len(dm.drug_encoder.alphabet_dict), "hidden_dim": 64},
        "img": {"stem_downsample": 2, "kernel_size": 5, "hidden_dim": 32, "depth": 2},
    }

    if args.ckpt_path != "None":
        model = AMRModel.load_from_checkpoint(args.ckpt_path)
    else:
        model = AMRModel(
            spectrum_embedder="mlp",
            drug_embedder=args.drug_embedder,
            spectrum_kwargs={
                "n_inputs": 6000,
                "n_outputs": 64,
                "layer_dims": size_to_layer_dims[args.spectrum_embedder],
                "dropout": 0.2,
            },
            drug_kwargs=drug_embedder_kwargs[args.drug_embedder],
            lr=args.lr,
            weight_decay=0,
            lr_decay_factor=1,
            warmup_steps=250,
            scaled_dot_product=True,
        )

    val_ckpt = ModelCheckpoint(monitor="val_micro_auc", mode="max")
    callbacks = [
        val_ckpt,
        EarlyStopping(monitor="val_micro_auc", patience=10, mode="max"),
    ]
    logger = TensorBoardLogger(
        args.logs_path,
        name="finetunehosp_%s_%s_%s_%s_%s"
        % (
            args.drug_embedder,
            args.spectrum_embedder,
            args.lr,
            args.hospital,
            args.percent,
        ),
    )

    trainer = Trainer(
        accelerator="gpu",
        devices=args.devices,
        strategy="auto",
        max_epochs=250,
        callbacks=callbacks,
        logger=logger,
    )

    trainer.fit(model, dm.train_dataloader(), dm.val_dataloader())

    # validation
    p = trainer.predict(
        model, dataloaders=dm.val_dataloader(), ckpt_path=val_ckpt.best_model_path
    )

    trues = torch.cat([pp[0][:, 0] for pp in p]).numpy()
    preds = torch.cat([pp[0][:, 1] for pp in p]).numpy()
    locs = np.concatenate([pp[1] for pp in p])
    drug_names = np.concatenate([pp[2] for pp in p])

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\tval\t%.5f\t%.5f\t%.5f\t%.5f\n"
            % (
                val_ckpt.best_model_path,
                micro_roc_auc(trues, preds, locs, drug_names),
                macro_roc_auc(trues, preds, locs, drug_names),
                instance_roc_auc(trues, preds, locs, drug_names),
                prec_at_1_neg(trues, preds, locs, drug_names),
            )
        )

    # test
    p = trainer.predict(
        model, dataloaders=dm.test_dataloader(), ckpt_path=val_ckpt.best_model_path
    )

    trues = torch.cat([pp[0][:, 0] for pp in p]).numpy()
    preds = torch.cat([pp[0][:, 1] for pp in p]).numpy()
    locs = np.concatenate([pp[1] for pp in p])
    drug_names = np.concatenate([pp[2] for pp in p])

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\ttest\t%.5f\t%.5f\t%.5f\t%.5f\n"
            % (
                val_ckpt.best_model_path,
                micro_roc_auc(trues, preds, locs, drug_names),
                macro_roc_auc(trues, preds, locs, drug_names),
                instance_roc_auc(trues, preds, locs, drug_names),
                prec_at_1_neg(trues, preds, locs, drug_names),
            )
        )


if __name__ == "__main__":
    main()



import os

os.environ["OMP_NUM_THREADS"] = "6"  # export OMP_NUM_THREADS=1
os.environ["OPENBLAS_NUM_THREADS"] = "6"  # export OPENBLAS_NUM_THREADS=1
os.environ["MKL_NUM_THREADS"] = "6"  # export MKL_NUM_THREADS=1
os.environ["VECLIB_MAXIMUM_THREADS"] = "6"  # export VECLIB_MAXIMUM_THREADS=1
os.environ["NUMEXPR_NUM_THREADS"] = "6"  # export NUMEXPR_NUM_THREADS=1

import torch
import numpy as np
from maldi_nn.utils.data import SpeciesClfDataModule
import sys
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from torchmetrics.functional.classification import multiclass_accuracy
from sklearn.model_selection import ParameterGrid
import argparse

class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass

def read_data(path):
    dm = SpeciesClfDataModule(
        path,
        batch_size=128,
        n_workers=2,
        preprocessor=None,
        in_memory=True,
    )
    dm.setup(None)


    X_train = torch.stack([b["intensity"] for b in dm.train]).numpy()
    y_train = np.array([b["species"] for b in dm.train])
    X_val = torch.stack([b["intensity"] for b in dm.val]).numpy()
    y_val = np.array([b["species"] for b in dm.val])
    X_test = torch.stack([b["intensity"] for b in dm.test]).numpy()
    y_test = np.array([b["species"] for b in dm.test])
    test_locs = np.array([b["0/loc"] for b in dm.test])
    return X_train, X_val, X_test, y_train, y_val, y_test, test_locs

def fit_knn(X_train, y_train, params):
    if params["norm"] == "standardscaler":
        model = Pipeline(
            steps=[
                ("norm", StandardScaler()),
                (
                    "knn",
                    KNeighborsClassifier(
                        n_neighbors=params["n_neighbors"],
                    ),
                ),
            ]
        )
    else:
        model = KNeighborsClassifier(
                    n_neighbors=params["n_neighbors"],
                )
        
    model.fit(X_train, y_train)
    return model

def fit_lr(X_train, y_train, params):
    if params["norm"] == "standardscaler":
        model = Pipeline(
            steps=[
                ("norm", StandardScaler()),
                (
                    "lr",
                    LogisticRegression(
                        solver="lbfgs",
                        max_iter=500,
                        C=params["C"],
                    ),
                ),
            ]
        )
    else:
        model = LogisticRegression(
            solver="lbfgs",
            max_iter=500,
            C=params["C"],
        )
    model.fit(X_train, y_train)
    return model

def fit_rf(X_train, y_train, params):
    model = RandomForestClassifier(**params, n_estimators = 200, n_jobs = 12)
    model.fit(X_train, y_train)
    return model


def main_knn(args):
    X_train, X_val, X_test, y_train, y_val, y_test, test_locs = read_data(args.path)

    knn_grid = ParameterGrid(
        {
            "norm": ["passthrough", "standardscaler"],
            "n_neighbors": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 25],
        }
    )

    scores = []
    for params in knn_grid:
        model = fit_knn(X_train, y_train, params)
        preds = torch.tensor(model.predict_proba(X_val))

        print(params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro"), flush = True)
        scores.append(
            [params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro")]
        )

    max_index = np.argmax([j[-1] for j in scores])
    winning_params = scores[max_index][0]

    winning_model = fit_knn(X_train, y_train, winning_params)

    np.savez(
        args.outputs,
        **{
            "trues": y_test,
            "preds": winning_model.predict_proba(X_test),
            "locs": test_locs,
        }
    )

def main_lr(args):
    X_train, X_val, X_test, y_train, y_val, y_test, test_locs = read_data(args.path)
    lr_grid = ParameterGrid(
        {
            "norm": ["passthrough", "standardscaler"],
            "C": 10.0 ** np.arange(-3, 4),
        }
    )
    scores = []
    for params in lr_grid:
        model = fit_lr(X_train, y_train, params)
        preds = torch.tensor(model.predict_proba(X_val))

        print(params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro"), flush = True)
        scores.append(
            [params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro")]
        )

    max_index = np.argmax([j[-1] for j in scores])
    winning_params = scores[max_index][0]
    winning_model = fit_lr(X_train, y_train, winning_params)

    np.savez(
        outputs,
        **{
            "trues": y_test,
            "preds": winning_model.predict_proba(X_test),
            "locs": test_locs,
        }
    )

def main_rf(args):
    X_train, X_val, X_test, y_train, y_val, y_test, test_locs = read_data(args.path)

    rf_grid = ParameterGrid(
        {
            "max_depth" : [25, 50, 75, 100],
            "min_samples_split": [2, 5, 10],
            "max_features": [10, 25, 50, 100],
        }
    )

    scores = []
    for params in rf_grid:
        model = fit_rf(X_train, y_train, params)
        preds = torch.tensor(model.predict_proba(X_val))

        print(params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro"), flush = True)
        scores.append(
            [params, multiclass_accuracy(preds, torch.tensor(y_val), num_classes = preds.shape[1], average="micro")]
        )

    max_index = np.argmax([j[-1] for j in scores])
    winning_params = scores[max_index][0]

    winning_model = fit_rf(X_train, y_train, winning_params)

    np.savez(
        args.outputs,
        **{
            "trues": y_test,
            "preds": winning_model.predict_proba(X_test),
            "locs": test_locs,
        }
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Training script for species identification baselines. Returns an npz file with predictions for the test set.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument(
        "outputs",
        type=str,
        metavar="outputs.npz",
        help="numpy .npz file to write (test) predictions into.",
    )
    parser.add_argument(
        "modeltype",
        type=str,
        metavar="modeltype",
        choices=["knn", "lr", "rf"],
        help="Which modeltype to use as baseline, choices: {%(choices)s}",
    )

    args = parser.parse_args()

    if args.modeltype == "knn":
        main_knn(args)
    elif args.modeltype == "lr":
        main_lr(args)
    elif args.modeltype == "rf":
        main_rf(args)
    


from maldi_nn.utils.data import *
import sys
from importlib.resources import files
import h5torch
import json
import os
import pandas as pd
import numpy as np
from maldi_nn.spectrum import *
import shutil

# Obtained from original DRIAMS publication Weis et al.
drug_to_class = {
    "5-Fluorocytosine": "antimycotic systemic",
    "Amikacin": "aminoglycoside",
    "Amoxicillin": "penicillins",
    "Amoxicillin + Clavulanic acid": "penicillins",
    "Amphotericin B": "antimycotic systemic",
    "Ampicillin": "penicillins",
    "Ampicillin + Amoxicillin": "penicillins",
    "Ampicillin-Sulbactam": "penicillins",
    "Anidulafungin": "antimycotic systemic",
    "Azithromycin": "macrolides",
    "Aztreonam": "other beta-lactam",
    "Bacitracin A": "other",
    "Caspofungin": "antimycotic systemic",
    "Cefalotin + Cefazolin": "other beta-lactam",
    "Cefazolin": "other beta-lactam",
    "Cefepime": "other beta-lactam",
    "Cefixime": "other beta-lactam",
    "Cefotaxime": "other beta-lactam",
    "Cefoxitin": "other beta-lactam",
    "Cefpodoxime": "other beta-lactam",
    "Ceftaroline fosamil": "other beta-lactam",
    "Ceftazidime": "other beta-lactam",
    "Ceftazidime + Avibactam": "other beta-lactam",
    "Ceftobiprole": "other beta-lactam",
    "Ceftolozane + Tazobactam": "other beta-lactam",
    "Ceftriaxone": "other beta-lactam",
    "Cefuroxime": "other beta-lactam",
    "Chloramphenicol": "other",
    "Ciprofloxacin": "quinolone",
    "Clarithromycin": "macrolides",
    "Clindamycin": "macrolides",
    "Colistin": "other",
    "Cotrimoxazole": "other",
    "Daptomycin": "other",
    "Doxycycline": "tetracyclin",
    "Ertapenem": "other beta-lactam",
    "Erythromycin": "macrolides",
    "Fluconazole": "antimycotic systemic",
    "Fosfomycin": "other",
    "Fosfomycin Tromethamine": "other",
    "Fusidic acid": "other",
    "Gentamicin": "aminoglycoside",
    "Imipenem": "other beta-lactam",
    "Isavuconazole": "antimycotic systemic",
    "Itraconazole": "antimycotic systemic",
    "Levofloxacin": "quinolone",
    "Linezolid": "other",
    "Meropenem": "other beta-lactam",
    "Meropenem + Vaborbactam": "other beta-lactam",
    "Metronidazole": "other",
    "Micafungin": "antimycotic systemic",
    "Minocycline": "tetracyclin",
    "Moxifloxacin": "quinolone",
    "Mupirocin": "other",
    "Nitrofurantoin": "other",
    "Norfloxacin": "quinolone",
    "Novobiocin": "other",
    "Oxacillin": "penicillins",
    "Penicillin": "penicillins",
    "Piperacillin": "penicillins",
    "Piperacillin-Tazobactam": "penicillins",
    "Polymyxin B": "other",
    "Posaconazole": "antimycotic systemic",
    "Rifampicin": "other",
    "Streptomycin": "other",
    "Teicoplanin": "other",
    "Telithromycin": "macrolides",
    "Tetracycline": "tetracyclin",
    "Ticarcillin": "penicillins",
    "Ticarcillin-clavulanic Acid": "penicillins",
    "Tigecycline": "tetracyclin",
    "Tobramycin": "aminoglycoside",
    "Vancomycin": "other",
    "Voriconazole": "antimycotic systemic",
}


# Originally obtained via pubchempy
drug_to_smiles = {
    "Amoxicillin + Clavulanic acid": "CC1(C)SC2C(NC(=O)C(N)c3ccc(O)cc3)C(=O)N2C1C(=O)O.O=C(O)C1C(=CCO)OC2CC(=O)N21",
    "Ampicillin": "CC1(C)SC2C(NC(=O)C(N)c3ccccc3)C(=O)N2C1C(=O)O",
    "Ceftazidime": "CC(C)(ON=C(C(=O)NC1C(=O)N2C(C(=O)[O-])=C(C[n+]3ccccc3)CSC12)c1csc(N)n1)C(=O)O",
    "Ciprofloxacin": "O=C(O)c1cn(C2CC2)c2cc(N3CCNCC3)c(F)cc2c1=O",
    "Gentamicin": "CNC(C)C1CCC(N)C(OC2C(N)CC(N)C(OC3OCC(C)(O)C(NC)C3O)C2O)O1",
    "Ceftriaxone": "CON=C(C(=O)NC1C(=O)N2C(C(=O)O)=C(CSc3nc(=O)c(=O)[nH]n3C)CSC12)c1csc(N)n1",
    "Cefuroxime": "CON=C(C(=O)NC1C(=O)N2C(C(=O)O)=C(COC(N)=O)CSC12)c1ccco1",
    "Nitrofurantoin": "O=C1CN(N=Cc2ccc([N+](=O)[O-])o2)C(=O)N1",
    "Fosfomycin": "CC1OC1P(=O)(O)O",
    "Norfloxacin": "CCn1cc(C(=O)O)c(=O)c2cc(F)c(N3CCNCC3)cc21",
    "Polymyxin B": "CCC(C)CCCCC(=O)NC(CCN)C(=O)NC(C(=O)NC(CCN)C(=O)NC1CCNC(=O)C(C(C)O)NC(=O)C(CCN)NC(=O)C(CCN)NC(=O)C(CC(C)C)NC(=O)C(Cc2ccccc2)NC(=O)C(CCN)NC1=O)C(C)O",
    "Cotrimoxazole": "COc1cc(Cc2cnc(N)nc2N)cc(OC)c1OC.Cc1cc(NS(=O)(=O)c2ccc(N)cc2)no1",
    "Clarithromycin": "CCC1OC(=O)C(C)C(OC2CC(C)(OC)C(O)C(C)O2)C(C)C(OC2OC(C)CC(N(C)C)C2O)C(C)(OC)CC(C)C(=O)C(C)C(O)C1(C)O",
    "Clindamycin": "CCCC1CC(C(=O)NC(C(C)Cl)C2OC(SC)C(O)C(O)C2O)N(C)C1",
    "Doxycycline": "CC1c2cccc(O)c2C(O)=C2C(=O)C3(O)C(O)=C(C(N)=O)C(=O)C(N(C)C)C3C(O)C21",
    "Fusidic acid": "CC(=O)OC1CC2(C)C(CC(O)C3C4(C)CCC(O)C(C)C4CCC32C)C1=C(CCC=C(C)C)C(=O)O",
    "Oxacillin": "Cc1onc(-c2ccccc2)c1C(=O)NC1C(=O)N2C1SC(C)(C)C2C(=O)O",
    "Penicillin": "CC1(C)SC2C(NC(=O)Cc3ccccc3)C(=O)N2C1C(=O)O",
    "Rifampicin": "COC1C=COC2(C)Oc3c(C)c(O)c4c(O)c(c(C=NN5CCN(C)CC5)c(O)c4c3C2=O)NC(=O)C(C)=CC=CC(C)C(O)C(C)C(O)C(C)C(OC(C)=O)C1C",
    "Vancomycin": "CNC(CC(C)C)C(=O)NC1C(=O)NC(CC(N)=O)C(=O)NC2C(=O)NC3C(=O)NC(C(=O)NC(C(=O)O)c4cc(O)cc(O)c4-c4cc3ccc4O)C(O)c3ccc(c(Cl)c3)Oc3cc2cc(c3OC2OC(CO)C(O)C(O)C2OC2CC(C)(N)C(O)C(C)O2)Oc2ccc(cc2Cl)C1O",
    "Linezolid": "CC(=O)NCC1CN(c2ccc(N3CCOCC3)c(F)c2)C(=O)O1",
    "Mupirocin": "CC(=CC(=O)OCCCCCCCCC(=O)O)CC1OCC(CC2OC2C(C)C(C)O)C(O)C1O",
    "Piperacillin-Tazobactam": "CC1(Cn2ccnn2)C(C(=O)O)N2C(=O)CC2S1(=O)=O.CCN1CCN(C(=O)NC(C(=O)NC2C(=O)N3C2SC(C)(C)C3C(=O)[O-])c2ccccc2)C(=O)C1=O.[Na+]",
    "Metronidazole": "Cc1ncc([N+](=O)[O-])n1CCO",
    "Moxifloxacin": "COc1c(N2CC3CCCNC3C2)c(F)cc2c(=O)c(C(=O)O)cn(C3CC3)c12",
    "Amikacin": "NCCC(O)C(=O)NC1CC(N)C(OC2OC(CN)C(O)C(O)C2O)C(O)C1OC1OC(CO)C(O)C(N)C1O",
    "Cefepime": "CON=C(C(=O)NC1C(=O)N2C(C(=O)[O-])=C(C[N+]3(C)CCCC3)CSC12)c1csc(N)n1",
    "Imipenem": "CC(O)C1C(=O)N2C(C(=O)O)=C(SCCN=CN)CC12",
    "Azithromycin": "CCC1OC(=O)C(C)C(OC2CC(C)(OC)C(O)C(C)O2)C(C)C(OC2OC(C)CC(N(C)C)C2O)C(C)(O)CC(C)CN(C)C(C)C(O)C1(C)O",
    "Erythromycin": "CCC1OC(=O)C(C)C(OC2CC(C)(OC)C(O)C(C)O2)C(C)C(OC2OC(C)CC(N(C)C)C2O)C(C)(O)CC(C)C(=O)C(C)C(O)C1(C)O",
    "Cefalotin + Cefazolin": "CC(=O)OCC1=C(C(=O)O)N2C(=O)C(NC(=O)Cc3cccs3)C2SC1.Cc1nnc(SCC2=C(C(=O)O)N3C(=O)C(NC(=O)Cn4cnnn4)C3SC2)s1",
    "Tetracycline": "CN(C)C1C(=O)C(C(N)=O)=C(O)C2(O)C(=O)C3=C(O)c4c(O)cccc4C(C)(O)C3CC12",
    "Novobiocin": "COC1C(OC(N)=O)C(O)C(Oc2ccc3c(O)c(NC(=O)c4ccc(O)c(CC=C(C)C)c4)c(=O)oc3c2C)OC1(C)C",
    "Meropenem": "CC(O)C1C(=O)N2C(C(=O)O)=C(SC3CNC(C(=O)N(C)C)C3)C(C)C12",
    "Daptomycin": "CCCCCCCCCC(=O)NC(Cc1c[nH]c2ccccc12)C(=O)NC(CC(N)=O)C(=O)NC(CC(=O)O)C(=O)NC1C(=O)NCC(=O)NC(CCCN)C(=O)NC(CC(=O)O)C(=O)NC(C)C(=O)NC(CC(=O)O)C(=O)NCC(=O)NC(CO)C(=O)NC(C(C)CC(=O)O)C(=O)NC(CC(=O)c2ccccc2N)C(=O)OC1C",
    "5-Fluorocytosine": "Nc1[nH]c(=O)ncc1F",
    "Amphotericin B": "CC1C=CC=CC=CC=CC=CC=CC=CC(OC2OC(C)C(O)C(N)C2O)CC2OC(O)(CC(O)CC(O)C(O)CCC(O)CC(O)CC(=O)OC(C)C(C)C1O)CC(O)C2C(=O)O",
    "Amoxicillin": "CC1(C)SC2C(NC(=O)C(N)c3ccc(O)cc3)C(=O)N2C1C(=O)O",
    "Aztreonam": "CC1C(NC(=O)C(=NOC(C)(C)C(=O)O)c2csc(N)n2)C(=O)N1S(=O)(=O)O",
    "Caspofungin": "CCC(C)CC(C)CCCCCCCCC(=O)NC1CC(O)C(NCCN)NC(=O)C2C(O)CCN2C(=O)C(C(O)CCN)NC(=O)C(C(O)C(O)c2ccc(O)cc2)NC(=O)C2CC(O)CN2C(=O)C(C(C)O)NC1=O",
    "Ceftolozane + Tazobactam": "CC1(Cn2ccnn2)C(C(=O)O)N2C(=O)CC2S1(=O)=O.Cn1c(N)c(NC(=O)NCCN)c[n+]1CC1=C(C(=O)[O-])N2C(=O)C(NC(=O)C(=NOC(C)(C)C(=O)O)c3nsc(N)n3)C2SC1",
    "Colistin": "CCC(C)CCCC(=O)NC(CCN)C(=O)NC(C(=O)NC(CCN)C(=O)NC1CCNC(=O)C(C(C)O)NC(=O)C(CCN)NC(=O)C(CCN)NC(=O)C(CC(C)C)NC(=O)C(CC(C)C)NC(=O)C(CCN)NC1=O)C(C)O",
    "Cefotaxime": "CON=C(C(=O)NC1C(=O)N2C(C(=O)O)=C(COC(C)=O)CSC12)c1csc(N)n1",
    "Ertapenem": "CC(O)C1C(=O)N2C(C(=O)O)=C(SC3CNC(C(=O)Nc4cccc(C(=O)O)c4)C3)C(C)C12",
    "Fluconazole": "OC(Cn1cncn1)(Cn1cncn1)c1ccc(F)cc1F",
    "Cefoxitin": "COC1(NC(=O)Cc2cccs2)C(=O)N2C(C(=O)O)=C(COC(N)=O)CSC21",
    "Levofloxacin": "CC1COc2c(N3CCN(C)CC3)c(F)cc3c(=O)c(C(=O)O)cn1c23",
    "Minocycline": "CN(C)c1ccc(O)c2c1CC1CC3C(N(C)C)C(=O)C(C(N)=O)=C(O)C3(O)C(=O)C1=C2O",
    "Teicoplanin": "CCCCCCCCCC(=O)NC1C(Oc2c3cc4cc2Oc2ccc(cc2Cl)C(OC2OC(CO)C(O)C(O)C2NC(C)=O)C2NC(=O)C(NC(=O)C4NC(=O)C4NC(=O)C(Cc5ccc(c(Cl)c5)O3)NC(=O)C(N)c3ccc(O)c(c3)Oc3cc(O)cc4c3)c3ccc(O)c(c3)-c3c(OC4OC(CO)C(O)C(O)C4O)cc(O)cc3C(C(=O)O)NC2=O)OC(CO)C(O)C1O",
    "Tigecycline": "CN(C)c1cc(NC(=O)CNC(C)(C)C)c(O)c2c1CC1CC3C(N(C)C)C(=O)C(C(N)=O)=C(O)C3(O)C(=O)C1=C2O",
    "Tobramycin": "NCC1OC(OC2C(N)CC(N)C(OC3OC(CO)C(O)C(N)C3O)C2O)C(N)CC1O",
    "Voriconazole": "CC(c1ncncc1F)C(O)(Cn1cncn1)c1ccc(F)cc1F",
    "Ceftazidime + Avibactam": "CC(C)(ON=C(C(=O)NC1C(=O)N2C(C(=O)[O-])=C(C[n+]3ccccc3)CSC12)c1csc(N)n1)C(=O)O.NC(=O)C1CCC2CN1C(=O)N2OS(=O)(=O)O",
    "Meropenem + Vaborbactam": "CC(O)C1C(=O)N2C(C(=O)O)=C(SC3CNC(C(=O)N(C)C)C3)C(C)C12.O=C(O)CC1CCC(NC(=O)Cc2cccs2)B(O)O1",
    "Chloramphenicol": "O=C(NC(CO)C(O)c1ccc([N+](=O)[O-])cc1)C(Cl)Cl",
    "Cefpodoxime": "COCC1=C(C(=O)O)N2C(=O)C(NC(=O)C(=NOC)c3csc(N)n3)C2SC1",
    "Piperacillin": "CCN1CCN(C(=O)NC(C(=O)NC2C(=O)N3C2SC(C)(C)C3C(=O)O)c2ccccc2)C(=O)C1=O",
    "Ampicillin-Sulbactam": "CC1(C)C(C(=O)O)C2C(=O)CC2S1(=O)=O.CC1(C)SC2C(NC(=O)C(N)c3ccccc3)C(=O)N2C1C(=O)O",
    "Ticarcillin-clavulanic Acid": "CC1(C)SC2C(NC(=O)C(C(=O)O)c3ccsc3)C(=O)N2C1C(=O)O.O=C(O)C1C(=CCO)OC2CC(=O)N21",
    "Telithromycin": "CCC1OC(=O)C(C)C(=O)C(C)C(OC2OC(C)CC(N(C)C)C2O)C(C)(OC)CC(C)C(=O)C(C)C2N(CCCCn3cnc(-c4cccnc4)c3)C(=O)OC12C",
    "Ticarcillin": "CC1(C)SC2C(NC(=O)C(C(=O)O)c3ccsc3)C(=O)N2C1C(=O)O",
    "Streptomycin": "CNC1C(OC2C(OC3C(O)C(O)C(N=C(N)N)C(O)C3N=C(N)N)OC(C)C2(O)C=O)OC(CO)C(O)C1O",
    "Posaconazole": "CCC(C(C)O)n1ncn(-c2ccc(N3CCN(c4ccc(OCC5COC(Cn6cncn6)(c6ccc(F)cc6F)C5)cc4)CC3)cc2)c1=O",
    "Itraconazole": "CCC(C)n1ncn(-c2ccc(N3CCN(c4ccc(OCC5COC(Cn6cncn6)(c6ccc(Cl)cc6Cl)O5)cc4)CC3)cc2)c1=O",
    "Anidulafungin": "CCCCCOc1ccc(-c2ccc(-c3ccc(C(=O)NC4CC(O)C(O)NC(=O)C5C(O)C(C)CN5C(=O)C(C(C)O)NC(=O)C(C(O)C(O)c5ccc(O)cc5)NC(=O)C5CC(O)CN5C(=O)C(C(C)O)NC4=O)cc3)cc2)cc1",
    "Micafungin": "CCCCCOc1ccc(-c2cc(-c3ccc(C(=O)NC4CC(O)C(O)NC(=O)C5C(O)C(C)CN5C(=O)C(C(O)CC(N)=O)NC(=O)C(C(O)C(O)c5ccc(O)c(OS(=O)(=O)O)c5)NC(=O)C5CC(O)CN5C(=O)C(C(C)O)NC4=O)cc3)no2)cc1",
    "Ampicillin + Amoxicillin": "CC1(C)SC2C(NC(=O)C(N)c3ccccc3)C(=O)N2C1C(=O)O.CC1(C)SC2C(NC(=O)C(N)c3ccc(O)cc3)C(=O)N2C1C(=O)O",
    "Fosfomycin Tromethamine": "CC1OC1P(=O)(O)O.NC(CO)(CO)CO",
    "Cefazolin": "Cc1nnc(SCC2=C(C(=O)O)N3C(=O)C(NC(=O)Cn4cnnn4)C3SC2)s1",
    "Cefixime": "C=CC1=C(C(=O)O)N2C(=O)C(NC(=O)C(=NOCC(=O)O)c3csc(N)n3)C2SC1",
    "Ceftaroline fosamil": "CCON=C(C(=O)NC1C(=O)N2C(C(=O)[O-])=C(Sc3nc(-c4cc[n+](C)cc4)cs3)CSC12)c1nsc(NP(=O)(O)O)n1",
    "Ceftobiprole": "Nc1nc(C(=NO)C(=O)NC2C(=O)N3C(C(=O)O)=C(C=C4CCN(C5CCNC5)C4=O)CSC23)ns1",
    "Bacitracin A": "CCC(C)C(N)C1=NC(C(=O)NC(CC(C)C)C(=O)NC(CCC(=O)O)C(=O)NC(C(=O)NC2CCCCNC(=O)C(CC(N)=O)NC(=O)C(CC(=O)O)NC(=O)C(Cc3cnc[nH]3)NC(=O)C(Cc3ccccc3)NC(=O)C(C(C)CC)NC(=O)C(CCCN)NC2=O)C(C)CC)CS1",
    "Isavuconazole": "CC(c1nc(-c2ccc(C#N)cc2)cs1)C(O)(Cn1cncn1)c1cc(F)ccc1F",
}


def DRIAMS_raw_spectra_to_h5torch(DRIAMS_ROOT, outfile):
    print("(1) gathering all spectra files ...")
    ids = []
    for ix, (root, dirs, files_) in enumerate(os.walk(DRIAMS_ROOT)):
        if "raw" in root:
            for f in files_:  # walk through all raw files
                if f.endswith(".txt"):
                    k = root + "/" + f
                    ids.append("/".join(k.split("/")[-4:]))

    print("(2) gathering species data for all spectra files ...")
    alldata = []
    for root, dirs, files_ in os.walk(DRIAMS_ROOT):
        if "id" in root:
            for f in files_:
                if f.endswith("_clean.csv"):
                    print("... reading amr data for " + root + "/" + f)
                    k = root + "/" + f
                    data = pd.read_csv(k)

                    data["code"] = (
                        k.split("/")[-4]
                        + "/raw/"
                        + k.split("/")[-2]
                        + "/"
                        + data["code"]
                        + ".txt"
                    )
                    ids_driams_year = [
                        f
                        for f in ids
                        if (f.startswith(k.split("/")[-4]))
                        and (k.split("/")[-2] == f.split("/")[2])
                    ]
                    # only keep amr data for files that exist
                    keepers = [d in ids_driams_year for d in data["code"]]
                    alldata.append(data.iloc[keepers, :])

    # cleanup data
    alldata = pd.concat(alldata).drop(
        ["laboratory_species", "Unnamed: 0.1", "Unnamed: 0", "combined_code", "genus"],
        axis=1,
    )[["code", "species"]]

    # putting unidentified species to nans
    alldata.loc[
        alldata["species"].str.contains("not reliable identification"), ["species"]
    ] = np.nan

    alldata = pd.concat(
        [
            pd.DataFrame({"code": list(set(ids).difference(set(alldata["code"])))}),
            alldata,
        ]
    )
    dataraw = alldata

    species_labels = np.unique(dataraw["species"].astype(bytes))
    _, species = np.where(
        dataraw["species"].astype(bytes).values.reshape(-1, 1) == species_labels
    )

    loc = dataraw["code"].values.astype(bytes)

    data_path = files("maldi_nn.utils").joinpath("driams_split.json")
    split = json.load(open(data_path))

    spectrum_split = np.array(
        [
            split[l]
            for l in pd.Series(loc.astype(str))
            .str.split("/", expand=True)[[0, 2, 3]]
            .apply("/".join, axis=1)
            .values
        ]
    )

    f = h5torch.File(outfile, "w")

    f.register(species, "central")
    f.register(loc, 0, name="loc")
    f.register(species_labels, "unstructured", name="species_labels")
    f.register(spectrum_split.astype(bytes), "unstructured", name="split")

    ints = []
    mzs = []
    for ix, k in enumerate(loc):
        spectrum = pd.read_table(
            os.path.join(DRIAMS_ROOT, k.astype(str)),
            comment="#",
            sep=" ",
            index_col=None,
            header=0,
        ).values
        spectrum = spectrum[~np.isnan(spectrum).any(1)]

        mz = spectrum[:, 0]
        intensities = spectrum[:, 1]
        ints.append(intensities.astype(np.uint32))
        mzs.append(mz.astype(np.float32))
        if (ix + 1) % 1000 == 0:
            print(ix, end=" ", flush=True)
            if (ix + 1) == 1000:
                f.register(ints, 0, name="intensity", mode="vlen", length=len(species))
                f.register(mzs, 0, name="mz", mode="vlen", length=len(species))
                ints = []
                mzs = []
            else:
                f.append(ints, "0/intensity")
                f.append(mzs, "0/mz")
                ints = []
                mzs = []

    f.append(ints, "0/intensity")
    f.append(mzs, "0/mz")
    f.close()
    print("done")
    return None


def DRIAMS_raw_amr_to_h5torch(DRIAMS_ROOT, outfile):
    print("(1) gathering all spectra files ...")
    ids = []
    for ix, (root, dirs, files_) in enumerate(os.walk(DRIAMS_ROOT)):
        if "raw" in root:
            for f in files_:  # walk through all raw files
                if f.endswith(".txt"):
                    k = root + "/" + f
                    ids.append("/".join(k.split("/")[-4:]))

    print("(2) gathering amr data for all spectra files ...")
    alldata = []
    for root, dirs, files_ in os.walk(DRIAMS_ROOT):
        if "id" in root:
            for f in files_:
                if f.endswith("_clean.csv"):
                    print("... reading amr data for " + root + "/" + f)
                    k = root + "/" + f
                    data = pd.read_csv(k)

                    data["code"] = (
                        k.split("/")[-4]
                        + "/raw/"
                        + k.split("/")[-2]
                        + "/"
                        + data["code"]
                        + ".txt"
                    )
                    ids_driams_year = [
                        f
                        for f in ids
                        if (f.startswith(k.split("/")[-4]))
                        and (k.split("/")[-2] == f.split("/")[2])
                    ]
                    # only keep amr data for files that exist
                    keepers = [d in ids_driams_year for d in data["code"]]
                    alldata.append(data.iloc[keepers, :])

    # cleanup data
    alldata_ = pd.concat(alldata).drop(
        ["laboratory_species", "Unnamed: 0.1", "Unnamed: 0", "combined_code", "genus"],
        axis=1,
    )

    # delete all labels that are not I, R or S
    alldata_without = alldata_.drop(columns=["code", "species"])
    alldata_without[
        ~(
            (alldata_without == "R")
            | (alldata_without == "I")
            | (alldata_without == "S")
        )
    ] = np.nan
    alldata = pd.concat([alldata_[["code", "species"]], alldata_without], axis=1)
    alldata = alldata.loc[
        :, alldata.nunique(dropna=False) != 1
    ]  # delete columns without labels

    # renaming and merging columns that have identical chemical structures
    def merge(data, names):
        data[names[0]] = (
            data[names]
            .astype(str)
            .apply("_".join, axis=1)
            .map(
                {
                    "R_nan": "R",
                    "nan_R": "R",
                    "S_nan": "S",
                    "nan_S": "S",
                    "I_nan": "I",
                    "nan_I": "I",
                    "R_R": "R",
                    "S_S": "S",
                    "I_I": "I",
                }
            )
        )
        data = data.drop([names[1]], axis=1)
        return data

    alldata = alldata.drop(["Quinolones", "Aminoglycosides"], axis=1)

    alldata = merge(alldata, ["Levofloxacin", "Ofloxacin"])  # are stereoisomers

    alldata = merge(alldata, ["Benzylpenicillin", "Benzylpenicillin_others"])
    alldata = merge(alldata, ["Benzylpenicillin", "Benzylpenicillin_with_meningitis"])
    alldata = merge(alldata, ["Benzylpenicillin", "Benzylpenicillin_with_pneumonia"])

    alldata = merge(alldata, ["Penicillin", "Penicillin_with_endokarditis"])
    alldata = merge(alldata, ["Penicillin", "Penicillin_without_meningitis"])
    alldata = merge(alldata, ["Penicillin", "Penicillin_without_endokarditis"])
    alldata = merge(alldata, ["Penicillin", "Penicillin_with_pneumonia"])
    alldata = merge(alldata, ["Penicillin", "Penicillin_with_meningitis"])
    alldata = merge(alldata, ["Penicillin", "Penicillin_with_other_infections"])

    alldata = merge(alldata, ["Penicillin", "Benzylpenicillin"])

    alldata = merge(alldata, ["Cefuroxime", "Cefuroxime.1"])
    alldata = merge(alldata, ["Cotrimoxazole", "Cotrimoxazol"])
    alldata = merge(alldata, ["Gentamicin", "Gentamicin_high_level"])
    alldata = merge(alldata, ["Cefoxitin", "Cefoxitin_screen"])
    alldata = merge(alldata, ["Teicoplanin", "Teicoplanin_GRD"])
    alldata = merge(alldata, ["Vancomycin", "Vancomycin_GRD"])
    alldata = merge(alldata, ["Rifampicin", "Rifampicin_1mg-l"])

    alldata = merge(alldata, ["Meropenem", "Meropenem_with_meningitis"])
    alldata = merge(alldata, ["Meropenem", "Meropenem_without_meningitis"])
    alldata = merge(alldata, ["Meropenem", "Meropenem_with_pneumonia"])

    alldata = merge(
        alldata,
        [
            "Amoxicillin-Clavulanic acid",
            "Amoxicillin-Clavulanic acid_uncomplicated_HWI",
        ],
    )

    # renaming scheme
    alldata.rename(
        columns={
            "Strepomycin_high_level": "Streptomycin",
            "Bacitracin": "Bacitracin A",
            "Ceftarolin": "Ceftaroline fosamil",
            "Fosfomycin-Trometamol": "Fosfomycin Tromethamine",
            "Amoxicillin-Clavulanic acid": "Amoxicillin + Clavulanic acid",
            "Ceftolozane-Tazobactam": "Ceftolozane + Tazobactam",
            "Ceftazidime-Avibactam": "Ceftazidime + Avibactam",
            "Meropenem-Vaborbactam": "Meropenem + Vaborbactam",
            "Ticarcillin-Clavulan acid": "Ticarcillin-clavulanic Acid",
            "Ampicillin-Amoxicillin": "Ampicillin + Amoxicillin",
            "Cefalotin-Cefazolin": "Cefalotin + Cefazolin",
        },
        inplace=True,
    )

    # putting unidentified species to nans
    alldata.loc[
        alldata["species"].str.contains("not reliable identification"), ["species"]
    ] = np.nan
    alldata = pd.concat(
        [
            pd.DataFrame({"code": list(set(ids).difference(set(alldata["code"])))}),
            alldata,
        ]
    )
    dataraw = alldata

    dataraw = dataraw[(~pd.isnull(dataraw.iloc[:, 2:])).sum(1) != 0]

    species_labels = np.unique(dataraw["species"].astype(bytes))
    _, species = np.where(
        dataraw["species"].astype(bytes).values.reshape(-1, 1) == species_labels
    )

    loc = dataraw["code"].values.astype(bytes)
    amr_row, amr_col = np.where(~pd.isnull(dataraw.iloc[:, 2:]))
    amr_val = dataraw.iloc[:, 2:].values[amr_row, amr_col].astype(bytes)
    amr_names = dataraw.columns[2:].values.astype(bytes)
    amr_classes = [drug_to_class[k.astype(str)] for k in amr_names]
    amr_smiles = [drug_to_smiles[k.astype(str)] for k in amr_names]

    data_path = files("maldi_nn.utils").joinpath("driams_split.json")
    split = json.load(open(data_path))

    spectrum_split = np.array(
        [
            split[l]
            for l in pd.Series(loc.astype(str))
            .str.split("/", expand=True)[[0, 2, 3]]
            .apply("/".join, axis=1)
            .values
        ]
    )

    amr_split = spectrum_split[amr_row]
    indices = np.stack([amr_row, amr_col])
    shp = list(indices.max(1) + 1)

    f = h5torch.File(outfile, "w")

    f.register((indices, amr_val, shp), "central", mode="coo")
    f.register(species, 0, name="species")
    f.register(loc, 0, name="loc")
    f.register(species_labels, "unstructured", name="species_labels")
    f.register(np.array(amr_smiles).astype(bytes), 1, name="drug_smiles")
    f.register(np.array(amr_classes).astype(bytes), 1, name="drug_classes")
    f.register(amr_names.astype(bytes), 1, name="drug_names")
    f.register(amr_split.astype(bytes), "unstructured", name="split")

    ints = []
    mzs = []
    for ix, k in enumerate(loc):
        spectrum = pd.read_table(
            os.path.join(DRIAMS_ROOT, k.astype(str)),
            comment="#",
            sep=" ",
            index_col=None,
            header=0,
        ).values
        spectrum = spectrum[~np.isnan(spectrum).any(1)]

        mz = spectrum[:, 0]
        intensities = spectrum[:, 1]
        ints.append(intensities.astype(np.uint32))
        mzs.append(mz.astype(np.float32))
        if (ix + 1) % 1000 == 0:
            print(ix, end=" ", flush=True)
            if (ix + 1) == 1000:
                f.register(ints, 0, name="intensity", mode="vlen", length=len(species))
                f.register(mzs, 0, name="mz", mode="vlen", length=len(species))
                ints = []
                mzs = []
            else:
                f.append(ints, "0/intensity")
                f.append(mzs, "0/mz")
                ints = []
                mzs = []

    f.append(ints, "0/intensity")
    f.append(mzs, "0/mz")
    f.close()

    print("done")
    return None


def DRIAMS_raw_to_binned(rawfile, processed_file):
    binner = SequentialPreprocessor(
        VarStabilizer(method="sqrt"),
        Smoother(halfwindow=10),
        BaselineCorrecter(method="SNIP", snip_n_iter=20),
        Trimmer(),
        Binner(),
        Normalizer(sum=1),
    )
    shutil.copy(rawfile, processed_file)
    file = h5torch.File(processed_file, "a")
    len_ = file["0/mz"].shape[0]
    ints = []
    for i in range(len_):
        mz = file["0/mz"][i]
        intensity = file["0/intensity"][i]
        s = SpectrumObject(mz=mz, intensity=intensity)
        ints.append(binner(s).intensity)
        if (i + 1) % 1000 == 0:
            print(i, end=" ", flush=True)

    del file["0/mz"]
    del file["0/intensity"]

    file.register(np.stack(ints), 0, name="intensity")
    file.register(binner(s).mz, "unstructured", name="mz")

    file.close()
    return None


def DRIAMS_raw_to_peaks(rawfile, processed_file):
    peakdetector = SequentialPreprocessor(
        VarStabilizer(method="sqrt"),
        Smoother(halfwindow=10),
        BaselineCorrecter(method="SNIP", snip_n_iter=20),
        Trimmer(),
        PersistenceTransformer(extract_nonzero=True),
        Normalizer(sum=1),
        PeakFilter(max_number=2048),
    )
    shutil.copy(rawfile, processed_file)
    file = h5torch.File(processed_file, "a")
    len_ = file["0/mz"].shape[0]
    ints = []
    mzs = []
    for i in range(len_):
        mz = file["0/mz"][i]
        intensity = file["0/intensity"][i]
        s = SpectrumObject(mz=mz, intensity=intensity)
        s = peakdetector(s)
        ints.append(s.intensity)
        mzs.append(s.mz)
        if (i + 1) % 1000 == 0:
            print(i, end=" ", flush=True)

    del file["0/mz"]
    del file["0/intensity"]
    file.register(ints, 0, name="intensity", mode="vlen")
    file.register(mzs, 0, name="mz", mode="vlen")

    file.close()
    return None


def main():
    DRIAMS_root = str(sys.argv[1])
    amrraw = str(sys.argv[2])
    spectraraw = str(sys.argv[3])
    amrbin = str(sys.argv[4])
    spectrabin = str(sys.argv[5])
    amrpks = str(sys.argv[6])
    spectrapks = str(sys.argv[7])

    DRIAMS_raw_amr_to_h5torch(DRIAMS_root, amrraw)
    DRIAMS_raw_spectra_to_h5torch(DRIAMS_root, spectraraw)

    DRIAMS_raw_to_binned(amrraw, amrbin)
    DRIAMS_raw_to_peaks(amrraw, amrpks)
    DRIAMS_raw_to_binned(spectraraw, spectrabin)
    DRIAMS_raw_to_peaks(spectraraw, spectrapks)


if __name__ == "__main__":
    main()



from scipy.signal import savgol_filter
from scipy import sparse
from scipy.linalg import norm
import pandas as pd
import numpy as np
from scipy.stats import binned_statistic
from maldi_nn.utils import topf
import matplotlib.pyplot as plt
import h5torch
import torch


class SpectrumObject:
    """Base Spectrum Object class

    Can be instantiated directly with 1-D np.arrays for mz and intensity.
    Alternatively, can be read from csv files or from bruker output data.
    Reading from Bruker data is based on the code in https://github.com/sgibb/readBrukerFlexData

    Parameters
    ----------
    mz : 1-D np.array, optional
        mz values, by default None
    intensity : 1-D np.array, optional
        intensity values, by default None
    """

    def __init__(self, mz=None, intensity=None):
        self.mz = mz
        self.intensity = intensity
        if self.intensity is not None:
            if np.issubdtype(self.intensity.dtype, np.unsignedinteger):
                self.intensity = self.intensity.astype(int)
        if self.mz is not None:
            if np.issubdtype(self.mz.dtype, np.unsignedinteger):
                self.mz = self.mz.astype(int)

    def __getitem__(self, index):
        return SpectrumObject(mz=self.mz[index], intensity=self.intensity[index])

    def __len__(self):
        if self.mz is not None:
            return self.mz.shape[0]
        else:
            return 0

    def plot(self, as_peaks=False):
        """Plot a spectrum via matplotlib

        Parameters
        ----------
        as_peaks : bool, optional
            draw points in the spectrum as individualpeaks, instead of connecting the points in the spectrum, by default False
        """
        if as_peaks:
            mz_plot = np.stack([self.mz - 1, self.mz, self.mz + 1]).T.reshape(-1)
            int_plot = np.stack(
                [
                    np.zeros_like(self.intensity),
                    self.intensity,
                    np.zeros_like(self.intensity),
                ]
            ).T.reshape(-1)
        else:
            mz_plot, int_plot = self.mz, self.intensity
        plt.plot(mz_plot, int_plot)

    def __repr__(self):
        string_ = np.array2string(
            np.stack([self.mz, self.intensity]), precision=5, threshold=10, edgeitems=2
        )
        mz_string, int_string = string_.split("\n")
        mz_string = mz_string[1:]
        int_string = int_string[1:-1]
        return "SpectrumObject([\n\tmz  = %s,\n\tint = %s\n])" % (mz_string, int_string)

    @staticmethod
    def tof2mass(ML1, ML2, ML3, TOF):
        A = ML3
        B = np.sqrt(1e12 / ML1)
        C = ML2 - TOF

        if A == 0:
            return (C * C) / (B * B)
        else:
            return ((-B + np.sqrt((B * B) - (4 * A * C))) / (2 * A)) ** 2

    @classmethod
    def from_bruker(cls, acqu_file, fid_file):
        """Read a spectrum from Bruker's format

        Parameters
        ----------
        acqu_file : str
            "acqu" file bruker folder
        fid_file : str
            "fid" file in bruker folder

        Returns
        -------
        SpectrumObject
        """
        with open(acqu_file, "rb") as f:
            lines = [line.decode("utf-8", errors="replace").rstrip() for line in f]
        for l in lines:
            if l.startswith("##$TD"):
                TD = int(l.split("= ")[1])
            if l.startswith("##$DELAY"):
                DELAY = int(l.split("= ")[1])
            if l.startswith("##$DW"):
                DW = float(l.split("= ")[1])
            if l.startswith("##$ML1"):
                ML1 = float(l.split("= ")[1])
            if l.startswith("##$ML2"):
                ML2 = float(l.split("= ")[1])
            if l.startswith("##$ML3"):
                ML3 = float(l.split("= ")[1])
            if l.startswith("##$BYTORDA"):
                BYTORDA = int(l.split("= ")[1])
            if l.startswith("##$NTBCal"):
                NTBCal = l.split("= ")[1]

        intensity = np.fromfile(fid_file, dtype={0: "<i", 1: ">i"}[BYTORDA])

        if len(intensity) < TD:
            TD = len(intensity)
        TOF = DELAY + np.arange(TD) * DW

        mass = cls.tof2mass(ML1, ML2, ML3, TOF)

        intensity[intensity < 0] = 0

        return cls(mz=mass, intensity=intensity)

    @classmethod
    def from_tsv(cls, file, sep=" "):
        """Read a spectrum from txt

        Parameters
        ----------
        file : str
            path to csv file
        sep : str, optional
            separator in the file, by default " "

        Returns
        -------
        SpectrumObject
        """
        s = pd.read_table(
            file, sep=sep, index_col=None, comment="#", header=None
        ).values
        mz = s[:, 0]
        intensity = s[:, 1]
        return cls(mz=mz, intensity=intensity)

    def torch(self):
        """Converts spectrum to dict of tensors"""
        return {"mz": torch.tensor(self.mz), "intensity": torch.tensor(self.intensity)}


class Binner:
    """Pre-processing function for binning spectra in equal-width bins.

    Parameters
    ----------
    start : int, optional
        start of the binning range, by default 2000
    stop : int, optional
        end of the binning range, by default 20000
    step : int, optional
        width of every bin, by default 3
    aggregation : str, optional
        how to aggregate intensity values in each bin.
        Is passed to the statistic argument of https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.binned_statistic.html
        Can take any argument that the statistic argument also takes, by default "sum"
    """

    def __init__(self, start=2000, stop=20000, step=3, aggregation="sum"):
        self.bins = np.arange(start, stop + 1e-8, step)
        self.mz_bins = self.bins[:-1] + step / 2
        self.agg = aggregation

    def __call__(self, SpectrumObj):
        if self.agg == "sum":
            bins, _ = np.histogram(
                SpectrumObj.mz, self.bins, weights=SpectrumObj.intensity
            )
        else:
            bins = binned_statistic(
                SpectrumObj.mz,
                SpectrumObj.intensity,
                bins=self.bins,
                statistic=self.agg,
            ).statistic
            bins = np.nan_to_num(bins)

        s = SpectrumObject(intensity=bins, mz=self.mz_bins)
        return s


class Normalizer:
    """Pre-processing function for normalizing the intensity of a spectrum.
    Commonly referred to as total ion current (TIC) calibration.

    Parameters
    ----------
    sum : int, optional
        Make the total intensity of the spectrum equal to this amount, by default 1
    """

    def __init__(self, sum=1):
        self.sum = sum

    def __call__(self, SpectrumObj):
        s = SpectrumObject()

        s = SpectrumObject(
            intensity=SpectrumObj.intensity / SpectrumObj.intensity.sum() * self.sum,
            mz=SpectrumObj.mz,
        )
        return s


class Trimmer:
    """Pre-processing function for trimming ends of a spectrum.
    This can be used to remove inaccurate measurements.

    Parameters
    ----------
    min : int, optional
        remove all measurements with mz's lower than this value, by default 2000
    max : int, optional
        remove all measurements with mz's higher than this value, by default 20000
    """

    def __init__(self, min=2000, max=20000):
        self.range = [min, max]

    def __call__(self, SpectrumObj):
        indices = (self.range[0] < SpectrumObj.mz) & (SpectrumObj.mz < self.range[1])

        s = SpectrumObject(
            intensity=SpectrumObj.intensity[indices], mz=SpectrumObj.mz[indices]
        )
        return s


class VarStabilizer:
    """Pre-processing function for manipulating intensities.
    Commonly performed to stabilize their variance.

    Parameters
    ----------
    method : str, optional
        function to apply to intensities.
        can be either "sqrt", "log", "log2" or "log10", by default "sqrt"
    """

    def __init__(self, method="sqrt"):
        methods = {"sqrt": np.sqrt, "log": np.log, "log2": np.log2, "log10": np.log10}
        self.fun = methods[method]

    def __call__(self, SpectrumObj):
        s = SpectrumObject(intensity=self.fun(SpectrumObj.intensity), mz=SpectrumObj.mz)
        return s


class BaselineCorrecter:
    """Pre-processing function for baseline correction (also referred to as background removal).

    Support SNIP, ALS and ArPLS.
    Some of the code is based on https://stackoverflow.com/questions/29156532/python-baseline-correction-library.

    Parameters
    ----------
    method : str, optional
        Which method to use
        either "SNIP", "ArPLS" or "ALS", by default None
    als_lam : float, optional
        lambda value for ALS and ArPLS, by default 1e8
    als_p : float, optional
        p value for ALS and ArPLS, by default 0.01
    als_max_iter : int, optional
        max iterations for ALS and ArPLS, by default 10
    als_tol : float, optional
        stopping tolerance for ALS and ArPLS, by default 1e-6
    snip_n_iter : int, optional
        iterations of SNIP, by default 10
    """

    def __init__(
        self,
        method=None,
        als_lam=1e8,
        als_p=0.01,
        als_max_iter=10,
        als_tol=1e-6,
        snip_n_iter=10,
    ):
        self.method = method
        self.lam = als_lam
        self.p = als_p
        self.max_iter = als_max_iter
        self.tol = als_tol
        self.n_iter = snip_n_iter

    def __call__(self, SpectrumObj):
        if "LS" in self.method:
            baseline = self.als(
                SpectrumObj.intensity,
                method=self.method,
                lam=self.lam,
                p=self.p,
                max_iter=self.max_iter,
                tol=self.tol,
            )
        elif self.method == "SNIP":
            baseline = self.snip(SpectrumObj.intensity, self.n_iter)

        s = SpectrumObject(
            intensity=SpectrumObj.intensity - baseline, mz=SpectrumObj.mz
        )
        return s

    def als(self, y, method="ArPLS", lam=1e8, p=0.01, max_iter=10, tol=1e-6):
        L = len(y)
        D = sparse.diags([1, -2, 1], [0, -1, -2], shape=(L, L - 2))
        D = lam * D.dot(
            D.transpose()
        )  # Precompute this term since it does not depend on `w`

        w = np.ones(L)
        W = sparse.spdiags(w, 0, L, L)

        crit = 1
        count = 0
        while crit > tol:
            z = sparse.linalg.spsolve(W + D, w * y)

            if method == "AsLS":
                w_new = p * (y > z) + (1 - p) * (y < z)
            elif method == "ArPLS":
                d = y - z
                dn = d[d < 0]
                m = np.mean(dn)
                s = np.std(dn)
                w_new = 1 / (1 + np.exp(np.minimum(2 * (d - (2 * s - m)) / s, 70)))

            crit = norm(w_new - w) / norm(w)
            w = w_new
            W.setdiag(w)
            count += 1
            if count > max_iter:
                break
        return z

    def snip(self, y, n_iter):
        y_prepr = np.log(np.log(np.sqrt(y + 1) + 1) + 1)
        for i in range(1, n_iter + 1):
            rolled = np.pad(y_prepr, (i, i), mode="edge")
            new = np.minimum(
                y_prepr, (np.roll(rolled, i) + np.roll(rolled, -i))[i:-i] / 2
            )
            y_prepr = new
        return (np.exp(np.exp(y_prepr) - 1) - 1) ** 2 - 1


class Smoother:
    """Pre-processing function for smoothing. Uses Savitzky-Golay filter.

    Parameters
    ----------
    halfwindow : int, optional
        halfwindow of savgol_filter, by default 10
    polyorder : int, optional
        polyorder of savgol_filter, by default 3
    """

    def __init__(self, halfwindow=10, polyorder=3):
        self.window = halfwindow * 2 + 1
        self.poly = polyorder

    def __call__(self, SpectrumObj):
        s = SpectrumObject(
            intensity=np.maximum(
                savgol_filter(SpectrumObj.intensity, self.window, self.poly), 0
            ),
            mz=SpectrumObj.mz,
        )
        return s


class PersistenceTransformer:
    """Pre-processing function for Peak Detection.
    Uses the Persistance Transformation first outlined in https://doi.org/10.1093/bioinformatics/btaa429
    Underlying code is from https://github.com/BorgwardtLab/Topf

    Parameters
    ----------
    extract_nonzero : bool, optional
        whether to extract detected peaks or to keep zeros in, by default False
    """

    def __init__(self, extract_nonzero=False):
        self.filter = extract_nonzero

    def __call__(self, SpectrumObj):
        a = np.stack([SpectrumObj.mz, SpectrumObj.intensity]).T
        b = topf.PersistenceTransformer().fit_transform(a)

        s = SpectrumObject()
        if self.filter:
            peaks = b[:, 1] != 0
            s = SpectrumObject(intensity=b[peaks, 1], mz=b[peaks, 0])
        else:
            s = SpectrumObject(intensity=b[:, 1], mz=b[:, 0])
        return s


class PeakFilter:
    """Pre-processing function for filtering peaks.

    Filters in two ways: absolute number of peaks and height.

    Parameters
    ----------
    max_number : int, optional
        Maximum number of peaks to keep. Prioritizes peaks to keep by height.
        by default None, for no filtering
    min_intensity : float, optional
        Min intensity of peaks to keep, by default None, for no filtering
    """

    def __init__(self, max_number=None, min_intensity=None):
        self.max_number = max_number
        self.min_intensity = min_intensity

    def __call__(self, SpectrumObj):
        s = SpectrumObject(intensity=SpectrumObj.intensity, mz=SpectrumObj.mz)

        if self.max_number is not None:
            indices = np.argsort(-s.intensity, kind="stable")
            take = np.sort(indices[: self.max_number])

            s.mz = s.mz[take]
            s.intensity = s.intensity[take]

        if self.min_intensity is not None:
            take = s.intensity >= self.min_intensity

            s.mz = s.mz[take]
            s.intensity = s.intensity[take]

        return s


class RandomPeakShifter:
    """Pre-processing function for adding random (gaussian) noise to the mz values of peaks.

    Parameters
    ----------
    std : float, optional
        stdev of the random noise to add, by default 1
    """

    def __init__(self, std=1.0):
        self.std = std

    def __call__(self, SpectrumObj):
        s = SpectrumObject(
            intensity=SpectrumObj.intensity,
            mz=SpectrumObj.mz
            + np.random.normal(scale=self.std, size=SpectrumObj.mz.shape),
        )
        return s


class UniformPeakShifter:
    """Pre-processing function for adding uniform noise to the mz values of peaks.

    Parameters
    ----------
    range : float, optional
        let each peak shift by maximum this value, by default 1.5
    """

    def __init__(self, range=1.5):
        self.range = range

    def __call__(self, SpectrumObj):
        s = SpectrumObject(
            intensity=SpectrumObj.intensity,
            mz=SpectrumObj.mz
            + np.random.uniform(
                low=-self.range, high=self.range, size=SpectrumObj.mz.shape
            ),
        )
        return s


class Binarizer:
    """Pre-processing function for binarizing intensity values of peaks.

    Parameters
    ----------
    threshold : float
        Threshold for the intensities to become 1 or 0.
    """

    def __init__(self, threshold):
        self.threshold = threshold

    def __call__(self, SpectrumObj):
        s = SpectrumObject(
            intensity=(SpectrumObj.intensity > self.threshold).astype(
                SpectrumObj.intensity.dtype
            ),
            mz=SpectrumObj.mz,
        )
        return s


class SequentialPreprocessor:
    """Chain multiple preprocessors so that a pre-processing pipeline can be called with one line.

    Example:
    ```python
    preprocessor = SequentialPreprocessor(
        VarStabilizer(),
        Smoother(),
        BaselineCorrecter(method="SNIP"),
        Normalizer(),
        Binner()
    )
    preprocessed_spectrum = preprocessor(spectrum)
    ```
    """

    def __init__(self, *args):
        self.preprocessors = args

    def __call__(self, SpectrumObj):
        for step in self.preprocessors:
            SpectrumObj = step(SpectrumObj)
        return SpectrumObj



import torch.nn as nn
import torch
from bio_attention.attention import TransformerEncoder
from bio_attention.embed import ContinuousEmbedding


class MLP(nn.Module):
    def __init__(
        self,
        n_inputs=6000,
        n_outputs=512,
        layer_dims=[512, 512, 512],
        layer_or_batchnorm="layer",
        dropout=0.2,
    ):
        super().__init__()

        c = n_inputs
        layers = []
        for i in layer_dims:
            layers.append(nn.Linear(c, i))
            layers.append(nn.GELU())
            layers.append(nn.Dropout(dropout))
            layers.append(
                nn.LayerNorm(i) if layer_or_batchnorm == "layer" else nn.BatchNorm1d(i)
            )
            c = i

        layers.append(nn.Linear(c, n_outputs))

        self.net = nn.Sequential(*layers)

        self.hsize = n_outputs

    def forward(self, spectrum):
        return self.net(spectrum["intensity"])


class Permute(nn.Module):
    def __init__(self, *args):
        super().__init__()
        self.args = args

    def forward(self, x):
        return x.permute(*self.args)


class Transformer(nn.Module):
    def __init__(
        self,
        depth,
        dim,
        n_heads=8,
        dropout=0.2,
        cls=True,
        reduce="none",
        output_head_dim=64,
    ):
        super().__init__()

        self.embed = ContinuousEmbedding(
            dim,
            depth=1,
            norm=False,
            cls=cls,
            init_cls_as_bias=True,
            init_mask_as_bias=True,
        )

        self.transformer = TransformerEncoder(
            depth=depth,
            dim=dim,
            nh=n_heads,
            attentiontype="vanilla",
            attention_args={"dropout": dropout},
            plugintype="sinusoidal",
            plugin_args={"dim": dim, "divide": 10},
            only_apply_plugin_at_first=True,
            dropout=dropout,
            glu_ff=True,
            activation="gelu",
        )

        self.reduce = reduce

        self.output_head = nn.Linear(dim, output_head_dim)

    def forward(self, spectrum):
        z = self.embed(spectrum["intensity"])
        z = self.transformer(z, pos=spectrum["mz"])

        if self.reduce == "mean":
            return self.output_head(z.sum(1))
        elif self.reduce == "max":
            return self.output_head(z.max(1).values)
        elif self.reduce == "cls":
            return self.output_head(z[:, 0, :])
        elif self.reduce == "none":
            return z



import os
import argparse
import numpy as np
import pandas as pd
import torch
from maldi_nn.models import SpeciesClassifier, MaldiTransformer
from maldi_nn.utils.data import SpeciesClfDataModule
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch import Trainer
from maldi_nn.spectrum import PeakFilter
import ast
from maldi_nn.reproduce.modules import (
    MaldiTransformerNegSampler,
    MaldiTransformerOnlyClf,
    MaldiTransformerMaskMSE
)

def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():
    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass


    parser = argparse.ArgumentParser(
        description="Training script for species identification.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument("logs_path", type=str, metavar="logs_path", help="path to logs.")

    parser.add_argument(
        "spectrum_embedder",
        type=str,
        metavar="spectrum_embedder",
        choices=["mlp", "trf"],
        help="Which spectrum embedder to use, choices: {%(choices)s}",
    )

    parser.add_argument(
        "size",
        type=str,
        metavar="size",
        choices=["Linear", "S", "M", "L", "XL"] ,
        help="Model size, choices: {%(choices)s}. Linear is only available for mlp.",
    )

    parser.add_argument("--lr", type=float, default=5e-4, help="Learning rate.")

    parser.add_argument(
        "--logging_file",
        type=str,
        default="res.txt",
        help="Which file to write final performances to.",
    )

    parser.add_argument(
        "--num_workers",
        type=int,
        default=4,
        help="Number of workers in dataloader. Reduce to alleviate CPU.",
    )

    parser.add_argument(
        "--devices",
        type=ast.literal_eval,
        default=1,
        help="devices to use. Input an integer to specify a number of gpus or a list e.g. [1] or [0,1,3] to specify which gpus.",
    )

    parser.add_argument(
        "--trf_n_peaks",
        type=int,
        default=200,
        help="Number of peaks",
    )

    parser.add_argument(
        "--trf_ckpt_path",
        type=str,
        default="None",
        help="Checkpoint path of malditransformer",
    )

    parser.add_argument(
        "--trf_ckpt_modeltype",
        type=str,
        default="vanilla",
        choices=["vanilla", "negpeaksampler", "intensitymlm", "onlyclf", "onlyshf"],
        help="Maldi Transformer pre-trained modeltype. choices: {%(choices)s}",
    )

    parser.add_argument(
        "--trf_transfer_output_head",
        type=boolean,
        default=False,
        help="Whether to transfer the clf output head of the Maldi Transformer. Can be set to true if domain adaptation was adopted.",
    )


    args = parser.parse_args()

    dm = SpeciesClfDataModule(
        args.path,
        batch_size=128,
        n_workers=args.num_workers,
        preprocessor=(
            None if args.spectrum_embedder == "mlp" else PeakFilter(max_number=args.trf_n_peaks)
        ),
        in_memory=True,
    )
    dm.setup(None)

    genus = pd.Series(dm.train.f["unstructured/species_labels"][:].view(np.ndarray)).astype(str).str.split(" ", expand=True)[0].values
    mapper = {k : v for v, k in enumerate(np.unique(genus))}
    genus_labels = torch.tensor([mapper[k] for k in genus])

    size_to_layer_dims = {
        "mlp": {
            "S": [256, 128],
            "M": [512, 256, 128],
            "L": [1024, 512, 256, 128],
            "XL": [2048, 1024, 512, 256, 128],
            "Linear": [],
        },
        "trf": {
            "S": [160, 4],
            "M": [184, 6],
            "L": [232, 8],
            "XL": [304, 10],
        },
    }

    if args.spectrum_embedder == "mlp":
        kwargs = {
            "n_inputs": 6000,
            "n_outputs": dm.n_species,
            "layer_dims": size_to_layer_dims["mlp"][args.size],
            "dropout": 0.2,
        }

    elif args.spectrum_embedder == "trf":
        kwargs = {
            "dim": size_to_layer_dims["trf"][args.size][0],
            "depth": size_to_layer_dims["trf"][args.size][1],
            "output_head_dim": dm.n_species,
            "reduce": "cls",
            "dropout": 0.2,
        }
        if args.trf_ckpt_path != "None":
            modeltype = MaldiTransformer
            if args.trf_ckpt_modeltype == "intensitymlm":
                modeltype = MaldiTransformerMaskMSE
            elif args.trf_ckpt_modeltype == "negpeaksampler":
                modeltype = MaldiTransformerNegSampler
            elif args.trf_ckpt_modeltype == "onlyclf":
                modeltype = MaldiTransformerOnlyClf

            malditransformer = modeltype.load_from_checkpoint(args.trf_ckpt_path)
            kwargs["depth"] = malditransformer.hparams.depth
            kwargs["dim"] = malditransformer.hparams.dim


    model = SpeciesClassifier(
        spectrum_embedder=args.spectrum_embedder,
        spectrum_kwargs=kwargs,
        lr=args.lr,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=250,
        genus_labels = genus_labels
    )

    if args.trf_ckpt_path != "None":
        pretrained_dict = malditransformer.transformer.state_dict()
        model_state_dict = model.spectrum_embedder.state_dict()
        if not args.trf_transfer_output_head:
            pretrained_dict = {
                k: v for k, v in pretrained_dict.items() if not k.startswith("output_head")
            }
        else:
            pretrained_dict = {
                k: v for k, v in pretrained_dict.items()
            }
        model_state_dict.update(pretrained_dict)
        model.spectrum_embedder.load_state_dict(model_state_dict)


    val_ckpt = ModelCheckpoint(monitor="val_acc", mode="max")
    callbacks = [val_ckpt, EarlyStopping(monitor="val_acc", patience=10, mode="max")]
    logger = TensorBoardLogger(
        args.logs_path,
        name="clf_%s_%s_%s_%s" % (args.spectrum_embedder, args.size, args.lr, args.trf_n_peaks),
    )

    trainer = Trainer(
        accelerator="gpu",
        devices=args.devices,
        strategy="auto",
        max_epochs=250,
        callbacks=callbacks,
        logger=logger,
    )

    trainer.fit(model, dm.train_dataloader(), dm.val_dataloader())

    # validation
    p = trainer.validate(
        model, dataloaders=dm.val_dataloader(), ckpt_path=val_ckpt.best_model_path
    )[0]

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\tval\t%.5f\t%.5f\n"
            % (val_ckpt.best_model_path, p["val_acc"], p["val_genus_acc"])
        )

    # test
    p = trainer.validate(
        model, dataloaders=dm.test_dataloader(), ckpt_path=val_ckpt.best_model_path
    )[0]

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\ttest\t%.5f\t%.5f\n"
            % (val_ckpt.best_model_path, p["val_acc"], p["val_genus_acc"])
        )

if __name__ == "__main__":
    main()


import argparse
import ast
from maldi_nn.models import MaldiTransformer
from maldi_nn.utils.data import SpeciesClfDataModule
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch import Trainer
from maldi_nn.spectrum import PeakFilter


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():
    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass


    parser = argparse.ArgumentParser(
        description="Training script for domain adaptation of a Maldi Transformer.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument("logs_path", type=str, metavar="logs_path", help="path to logs.")
    parser.add_argument(
        "spectrum_embedder",
        type=str,
        metavar="spectrum_embedder",
        choices=["S", "M", "L", "XL"],
        help="Which size spectrum embedder to use",
    )

    parser.add_argument(
        "--ckpt_path",
        type=str,
        default="None",
        help="Ckpt path",
    )

    parser.add_argument(
        "--n_peaks",
        type=int,
        default=200,
        help="Number of peaks",
    )

    parser.add_argument(
        "--p",
        type=float,
        default=0.5,
        help="shuffle freq",
    )

    parser.add_argument(
        "--lmbda",
        type=float,
        default=1 / 100,
        help="Lambda. This is the probability with which to apply the spec id loss per step. ",
    )

    parser.add_argument(
        "--lmbda2",
        type=float,
        default=1.,
        help="Additionally, fixed multiplier to apply to the spec id loss.",
    )
    parser.add_argument(
        "--steps",
        type=float,
        default=20_000,
        help="steps",
    )

    parser.add_argument(
        "--batch_size",
        type=int,
        default=512,
        help="batch size",
    )

    parser.add_argument("--lr", type=float, default=5e-4, help="Learning rate.")

    parser.add_argument(
        "--num_workers",
        type=int,
        default=4,
        help="Number of workers in dataloader. Reduce to alleviate CPU.",
    )
    parser.add_argument("--devices", type=ast.literal_eval)

    args = parser.parse_args()

    size_to_layer_dims = {
        "S": [160, 4],
        "M": [184, 6],
        "L": [232, 8],
        "XL": [304, 10],
    }

    dm = SpeciesClfDataModule(
        args.path,
        batch_size=args.batch_size,
        n_workers=args.num_workers,
        preprocessor=PeakFilter(max_number=args.n_peaks),
        in_memory=True,
    )
    dm.setup(None)


    model = MaldiTransformer(
        size_to_layer_dims[args.spectrum_embedder][1],
        size_to_layer_dims[args.spectrum_embedder][0],
        n_classes=dm.n_species,
        n_heads=8,
        dropout=0.2,
        p=args.p,
        clf=True,
        clf_train_p=args.lmbda,
        lmbda = args.lmbda2,
        lr=args.lr,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=2500,
        )

    if args.ckpt_path != "None":
        malditransformer = MaldiTransformer.load_from_checkpoint(args.ckpt_path)
        pretrained_dict = malditransformer.transformer.state_dict()
        model_state_dict = model.transformer.state_dict()
        pretrained_dict = {
            k: v for k, v in pretrained_dict.items() if not k.startswith("output_head")
        }
        model_state_dict.update(pretrained_dict)
        model.transformer.load_state_dict(model_state_dict)
        
        model.output_head.load_state_dict(malditransformer.output_head.state_dict())

    callbacks = [
        ModelCheckpoint(every_n_train_steps=5_000),
    ]
    logger = TensorBoardLogger(
        args.logs_path,
        name="domainadapt_%s_%s_%s_%s_%s"
        % (args.spectrum_embedder, args.n_peaks, args.p, args.lmbda, args.lr),
    )

    trainer = Trainer(
        accelerator="gpu",
        devices=args.devices,
        strategy="auto",
        gradient_clip_val=1,
        max_steps=args.steps,
        max_epochs=-1,
        val_check_interval=5_000,
        check_val_every_n_epoch=None,
        callbacks=callbacks,
        logger=logger,
        precision="bf16-mixed",
    )

    trainer.fit(model, dm)

if __name__ == "__main__":
    main()


"""`topf`: a package for topological peak filtering."""

# Copyright (c) 2018 Bastian Rieck
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__version__ = "0.1.0"

import numpy as np
import collections.abc


class UnionFind:
    """An implementation of a Union--Find class.

    Basic implementation of a Union--Find algorithm for a set of
    vertices with contiguous vertex indices. The class performs path
    compression by default and uses integers internally for storing
    a disjoint set.

    The class requires the vertices to form a contiguous sequence, so no
    gaps are allowed. The vertices also have to be zero-indexed.
    """

    def __init__(self, n_vertices):
        """Create new instance of the class.

        Creates a new Union--Find class with `n_vertices` vertices. The
        vertices are expected to be zero-indexed.

        Parameters
        ----------
        n_vertices : int
            Number of vertices
        """
        self.n_vertices = n_vertices
        self.parent = {x: x for x in range(n_vertices)}

    def find(self, u):
        """Find and return parent of `u`."""
        if self.parent[u] == u:
            return u
        else:
            # Perform path collapse operation
            self.parent[u] = self.find(self.parent[u])
            return self.parent[u]

    def merge(self, u, v):
        """Merge `u` into the component of `v`.

        Merges vertex u into the component of vertex v. Note the
        asymmetry of this operation.

        Parameters
        ----------
        u : int
            Source vertex
        v : int
            Target vertex
        """
        if u != v:
            self.parent[self.find(u)] = self.find(v)


class PersistenceDiagram(collections.abc.Sequence):
    """Persistence diagram storage class.

    Simple class for storing the pairs of a persistence diagram. This is
    nothing but a light-weight wrapper for additional convenience.
    """

    def __init__(self, pairs):
        """Create new diagram from sequence of pairs."""
        self._pairs = pairs

    def __len__(self):
        """Return number of persistence pairs."""
        return len(self._pairs)

    def __getitem__(self, index):
        """Return persistence at index `index`."""
        return self._pairs[index]

    def __str__(self):
        """Return string representation of the persistence diagram."""
        return str(self._pairs)

    def total_persistence(self, p=1.0):
        """Calculate total persistence of the persistence diagram.

        Calculates the sum of all persistence values in the diagram,
        weighted by the specified power.

        Parameters
        ----------
        p : float
            Exponent for the weight calculation. Must be positive, else
            an exception will be raised.

        Returns
        -------
        Total persistence of the persistence diagram.
        """
        assert p > 0.0

        persistence_values = self._pairs[:, 0] - self._pairs[:, 1]
        persistence_values = np.power(persistence_values, p)

        return np.power(np.sum(persistence_values), 1.0 / p)


class PersistenceTransformer:
    """Persistence transformer class.

    Transforms a function that is represented as an input array of (x,y)
    pairs into (x, persistence(x)) pairs. Optionally, all tuples of the
    corresponding persistence diagram are calculated as well.
    """

    def __init__(
        self,
        calculate_persistence_diagram=False,
        n_peaks=None,
        enforce_n_peaks=True,
    ):
        """Create new instance of the persistence transformer class.

        Parameters
        ----------
        calculate_persistence_diagram : bool
            If set, calculates the persistence diagram along with
            transformed function. Use the `persistence_diagram`
            property of the class to access it.

        n_peaks : bool
            If set, keeps only the specified number of peaks. Peaks will
            be eliminated in top-down order starting from the one with
            the lowest persistence. Thus, if `n_peaks` is 1, only the
            highest peak will be kept.

        enforce_n_peaks : bool
            If set, ensures that `n_peaks` peaks will be returned, even in
            cases where two peaks have the same persistence. This behaviour
            is on by default because if users request a certain number of
            peaks, they should get them. This follows the principle of least
            surprise.
        """
        self._calculate_persistence_diagram = calculate_persistence_diagram
        self._n_peaks = n_peaks
        self._persistence_diagram = None
        self._enforce_n_peaks = enforce_n_peaks

    def fit_transform(self, a):
        """Transform a single function according to its peaks.

        Parameters
        ----------
        a : array_like of shape (n, 2)
            Function with `n` 2D samples to be transformed into
            persistence space.

        Returns
        -------
        `np.array` of shape (n, 2), following the original input order
        of the function, but with each value having been transformed to
        a persistence value.
        """
        a = np.asarray(a)

        if len(a.shape) != 2 or a.shape[1] != 2:
            raise RuntimeError("Unexpected array format")

        # This way of sorting ensures that points with the same
        # y value will be sorted according to their x value. It
        # ensures that left-most points are detected first.
        indices = np.argsort(-a[:, 1], kind="stable")

        # Optionally, the function can also return a proper persistence
        # diagram, i.e. a set of tuples that describe the merges.
        if self._calculate_persistence_diagram:
            b = np.zeros_like(a)
            b[:, 0] = a[:, 1]  # y
            b[:, 1] = a[:, 1]  # y (everything is paired with itself)
        else:
            b = None

        # Prepare Union--Find data structure; by default, every vertex
        # is initialized to be its own parent.
        n_vertices = len(a)
        uf = UnionFind(n_vertices)

        # By default, all points that are not explicitly handled will be
        # assigned a persistence value of zero.
        persistence = np.zeros(n_vertices)

        for index in indices:
            left_index = index - 1
            right_index = index + 1

            y = a[index, 1]

            # Inner point: both neighbours are defined; this is easy to
            # handle because we just have to check both of them.
            if left_index >= 0 and right_index <= n_vertices - 1:
                y_left = a[left_index, 1]
                y_right = a[right_index, 1]

                # The point is a local minimum, so we have to merge the
                # two neighbours.
                if y_left >= y and y <= y_right:
                    # The left neighbour is the younger neighbour, so it
                    # will be merged into the right one.
                    if a[uf.find(left_index), 1] < a[uf.find(right_index), 1]:
                        persistence[uf.find(left_index)] = a[uf.find(left_index), 1] - y

                        if self._calculate_persistence_diagram:
                            b[uf.find(left_index), 1] = y

                        uf.merge(left_index, index)
                        uf.merge(index, right_index)
                    else:
                        persistence[uf.find(right_index)] = (
                            a[uf.find(right_index), 1] - y
                        )

                        if self._calculate_persistence_diagram:
                            b[uf.find(right_index), 1] = y

                        uf.merge(right_index, index)
                        uf.merge(index, left_index)

                # The point is a regular point, i.e. one neighbour
                # has a higher function value, the other one has a
                # lower function value.
                elif not (y > y_left and y > y_right):
                    # Always merge the current point into the higher one
                    # of its neighbours. This merge does not result in a
                    # pair.
                    if a[uf.find(left_index), 1] < a[uf.find(right_index), 1]:
                        uf.merge(index, right_index)
                    else:
                        uf.merge(index, left_index)

        # Assign the persistence value to the global maximum of the
        # function to ensure that all tuples have been paired.
        if n_vertices > 0:
            global_maximum_index = indices[0]
            global_minimum_index = indices[-1]

            persistence[global_maximum_index] = (
                a[global_maximum_index, 1] - a[global_minimum_index, 1]
            )

            if self._calculate_persistence_diagram:
                b[global_maximum_index, 1] = a[global_minimum_index, 1]

        # Only create a persistence diagram if we have some persistence
        # tuples to store.
        if b is not None:
            self._persistence_diagram = PersistenceDiagram(b)

        # Perform peak filtering: reduce the number of peaks such that
        # only `n_peaks` remain. If this is not possible (because of a
        # strange value distribution, try to approximate the number).
        if self._n_peaks is not None:
            persistence_values = sorted(persistence)[::-1]
            n_peaks = self._n_peaks

            assert n_vertices == len(persistence_values)

            # Error condition: no filtering should be done because the
            # number of peaks coincides with the number of points. The
            # client will not be notified here, because this condition
            # is only provided for readability.
            if n_peaks == len(persistence_values):
                pass

            # Error condition: there are fewer values than there are
            # peaks requested. In this case, we raise an error.
            elif n_peaks > len(persistence_values):
                raise RuntimeError(
                    f"Specified {n_peaks} peaks, but only {n_vertices} "
                    f"peaks are available."
                )

            # Perform the filtering
            threshold = persistence_values[n_peaks - 1]
            persistence[persistence < threshold] = 0

            # Duplicate values exist. Depending on our parameters, we
            # either ignore this or cut off the peaks that come later
            # in the sense of their x position.
            if persistence_values[n_peaks - 1] == persistence_values[n_peaks]:
                if self._enforce_n_peaks:
                    persistence[n_peaks + 1 :] = 0

        return np.vstack((a[:, 0], persistence)).T

    @property
    def persistence_diagram(self):
        """Return persistence diagram.

        Returns
        -------
        The persistence diagram that was optionally calculated by
        calling `fit_transform()`. The diagram, if available, will
        be returned as a 2D `numpy.array`.
        """
        return self._persistence_diagram



import torch
import h5torch
import numpy as np
from maldi_nn.utils.drug import (
    SmilesToIndex,
    SmilesToECFP,
    SmilesToAlphabet,
    SmilesToDEEPSmilesAlphabet,
    SmilesToSelfiesAlphabet,
    SmilesToImage,
    SmilesToLINGOKernel,
)
import re
from maldi_nn.spectrum import SpectrumObject
from lightning import LightningDataModule
from torch.nn.utils.rnn import pad_sequence
import pandas as pd


def spectra_to_torch(spectra):
    spectra = [s.torch() for s in spectra]
    return {
        "mz": torch.stack([s["mz"] for s in spectra]),
        "intensity": torch.stack([s["intensity"] for s in spectra]),
    }


class MALDITOFDataset(h5torch.Dataset):
    def __init__(self, path, subset=None, preprocessor=None):
        super().__init__(path, subset=subset, sample_processor=self.sample_processor)
        self.preprocessor = preprocessor

    @staticmethod
    def create(path, spectrumobjects, labels, binned=False, **objects):
        f = h5torch.File(path, "w")

        f.register(labels, "central")

        if binned:
            intensities = np.stack([s.intensity for s in spectrumobjects])
            mz = spectrumobjects[0].mz
            f.register(intensities, axis=0, name="intensity")
            f.register(mz, axis="unstructured", name="mz")
        else:
            intensities = [s.intensity for s in spectrumobjects]
            mzs = [s.mz for s in spectrumobjects]
            f.register(intensities, axis=0, name="intensity", mode="vlen")
            f.register(mzs, axis=0, name="mz", mode="vlen")

        for k, v in objects.items():
            f.register(v, axis=0, name=k)

    def sample_processor(self, f, sample):
        spectrum = SpectrumObject(
            mz=(sample["0/mz"] if "0/mz" in sample else f["unstructured/mz"][:]),
            intensity=sample["0/intensity"],
        )

        if self.preprocessor is not None:
            spectrum = self.preprocessor(spectrum)

        spectrum = {
            "intensity": torch.tensor(spectrum.intensity).float(),
            "mz": torch.tensor(spectrum.mz),
            "species" : sample["central"],
            "0/loc" : sample["0/loc"].astype(str)
        }

        sample = {k: v for k, v in sample.items() if k not in ["0/intensity", "0/mz", "central"]}
        return sample | spectrum


class MaldiDataModule(LightningDataModule):
    def __init__(self, batch_size=16, n_workers=4):
        super().__init__()
        self.n_workers = n_workers
        self.batch_size = batch_size

    def train_dataloader(self):
        return torch.utils.data.DataLoader(
            self.train,
            num_workers=self.n_workers,
            batch_size=self.batch_size,
            shuffle=True,
            pin_memory=True,
            collate_fn=batch_collater,
        )

    def val_dataloader(self):
        return torch.utils.data.DataLoader(
            self.val,
            num_workers=self.n_workers,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            collate_fn=batch_collater,
        )

    def test_dataloader(self):
        return torch.utils.data.DataLoader(
            self.test,
            num_workers=self.n_workers,
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
            collate_fn=batch_collater,
        )


class DRIAMSAMRDataModule(MaldiDataModule):
    def __init__(
        self,
        path,
        drug_encoder="onehot",
        drug_encoder_args={},  # only for ecfp and image
        batch_size=16,
        n_workers=4,
        preprocessor=None,
        min_spectrum_len=None,
        in_memory=False,
    ):
        super().__init__(batch_size=batch_size, n_workers=n_workers)
        self.path = path

        smiles_list = sorted(
            list(set(list(h5torch.File(path)["1/drug_smiles"][:].astype(str))))
        )
        if drug_encoder == "onehot":
            self.drug_encoder = SmilesToIndex(smiles_list)
        elif drug_encoder == "ecfp":
            self.drug_encoder = SmilesToECFP(smiles_list, **drug_encoder_args)
        elif drug_encoder == "img":
            self.drug_encoder = SmilesToImage(smiles_list, **drug_encoder_args)
        elif drug_encoder == "kernel":
            f = h5torch.File(path)
            trainmatcher = np.vectorize(lambda x: bool(re.match("A_train", x)))
            train_indices = np.where(
                trainmatcher(f["unstructured/split"][:].astype(str))
            )[0]
            cols = np.unique(f["central/indices"][:][1][train_indices])
            smiles_list = sorted(
                list(set(list(f["1/drug_smiles"][:][cols].astype(str))))
            )
            f.close()
            self.drug_encoder = SmilesToLINGOKernel(smiles_list)

        elif drug_encoder in ["cnn", "trf", "gru"]:
            if drug_encoder_args["alphabet"] == "smiles":
                self.drug_encoder = SmilesToAlphabet(smiles_list)
            if drug_encoder_args["alphabet"] == "deepsmiles":
                self.drug_encoder = SmilesToDEEPSmilesAlphabet(smiles_list)
            if drug_encoder_args["alphabet"] == "selfies":
                self.drug_encoder = SmilesToSelfiesAlphabet(smiles_list)

        self.min_spectrum_len = min_spectrum_len
        self.preprocessor = preprocessor
        self.in_memory = in_memory

    def setup(self, stage):
        f = h5torch.File(self.path)
        trainmatcher = np.vectorize(lambda x: bool(re.match("A_train", x)))
        train_indices = np.where(trainmatcher(f["unstructured/split"][:].astype(str)))[
            0
        ]
        valmatcher = np.vectorize(lambda x: bool(re.match("A_val", x)))
        val_indices = np.where(valmatcher(f["unstructured/split"][:].astype(str)))[0]
        testmatcher = np.vectorize(lambda x: bool(re.match("A_test", x)))
        test_indices = np.where(testmatcher(f["unstructured/split"][:].astype(str)))[0]

        if self.min_spectrum_len is not None:
            lens_all = np.array([len(tt) for tt in f["0/intensity"][:]])

            lens = lens_all[f["central"]["indices"][0][train_indices]]
            train_indices = train_indices[lens > self.min_spectrum_len]

            lens = lens_all[f["central"]["indices"][0][val_indices]]
            val_indices = val_indices[lens > self.min_spectrum_len]

            lens = lens_all[f["central"]["indices"][0][test_indices]]
            test_indices = test_indices[lens > self.min_spectrum_len]

        if self.in_memory:
            f = f.to_dict()

        self.train = h5torch.Dataset(
            f,
            sampling="coo",
            subset=train_indices,
            sample_processor=self.processor,
        )

        self.val = h5torch.Dataset(
            f,
            sampling="coo",
            subset=val_indices,
            sample_processor=self.processor,
        )

        self.test = h5torch.Dataset(
            f,
            sampling="coo",
            subset=test_indices,
            sample_processor=self.processor,
        )

    def processor(self, f, sample):
        spectrum = SpectrumObject(
            mz=(sample["0/mz"] if "0/mz" in sample else f["unstructured/mz"][:]),
            intensity=sample["0/intensity"],
        )

        if self.preprocessor is not None:
            spectrum = self.preprocessor(spectrum)
        spectrum = {
            "intensity": torch.tensor(spectrum.intensity).float(),
            "mz": torch.tensor(spectrum.mz),
        }

        return (
            spectrum
            | {"label": int(sample["central"].astype(str) != "S")}
            | {"drug": self.drug_encoder(sample["1/drug_smiles"].astype(str))}
            | {"species": sample["0/species"]}
            | {
                "loc": sample["0/loc"].astype(str),
                "drug_name": sample["1/drug_names"].astype(str),
            }
        )


class DRIAMSSpectrumDataModule(MaldiDataModule):
    def __init__(
        self,
        path,
        batch_size=512,
        n_workers=4,
        preprocessor=None,
        min_spectrum_len=None,
        in_memory=True,
        exclude_nans=False,
    ):
        super().__init__(batch_size=batch_size, n_workers=n_workers)
        self.path = path

        self.min_spectrum_len = min_spectrum_len

        f = h5torch.File(self.path)
        trainmatcher = np.vectorize(lambda x: bool(re.match("A_train", x)))
        self.train_indices = np.where(
            trainmatcher(f["unstructured/split"][:].astype(str))
        )[0]

        trainmatcher = np.vectorize(lambda x: bool(re.match("B_", x)))
        B_ = np.where(
            trainmatcher(f["unstructured/split"][:].astype(str))
        )[0]
        trainmatcher = np.vectorize(lambda x: bool(re.match("C_", x)))
        C_ = np.where(
            trainmatcher(f["unstructured/split"][:].astype(str))
        )[0]
        trainmatcher = np.vectorize(lambda x: bool(re.match("D_", x)))
        D_ = np.where(
            trainmatcher(f["unstructured/split"][:].astype(str))
        )[0]
        self.train_indices = np.concatenate([self.train_indices, B_, C_, D_])

        valmatcher = np.vectorize(lambda x: bool(re.match("A_val", x)))
        self.val_indices = np.where(valmatcher(f["unstructured/split"][:].astype(str)))[
            0
        ]
        testmatcher = np.vectorize(lambda x: bool(re.match("A_test", x)))
        self.test_indices = np.where(
            testmatcher(f["unstructured/split"][:].astype(str))
        )[0]

        if self.min_spectrum_len is not None:
            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.train_indices]])
            self.train_indices = self.train_indices[lens > self.min_spectrum_len]

            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.val_indices]])
            self.val_indices = self.val_indices[lens > self.min_spectrum_len]

            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.test_indices]])
            self.test_indices = self.test_indices[lens > self.min_spectrum_len]

        self.preprocessor = preprocessor
        self.in_memory = in_memory

        species_labels = f["unstructured/species_labels"][:]

        species_train = f["central"][:][self.train_indices]
        species_val = f["central"][:][self.val_indices]
        species_test = f["central"][:][self.test_indices]

        species_mapping = {i: i for i in range(len(species_labels))}

        nan = np.where(species_labels == b"nan")[0].item()

        mix_labels = pd.Series(species_labels.astype(str)).str.startswith("MIX").values
        for i in np.where(mix_labels)[0]:
            species_mapping[i] = nan

        s, c = np.unique(species_train, return_counts=True)
        for i in s[c < 5]:
            species_mapping[i] = nan

        exclude_val = np.array(
            list(set(np.unique(species_val)).difference(set(np.unique(species_train))))
        )
        exclude_test = np.array(
            list(set(np.unique(species_test)).difference(set(np.unique(species_train))))
        )
        for i in exclude_val:
            species_mapping[i] = nan
        for i in exclude_test:
            species_mapping[i] = nan

        # not in train/val/test of hosp A (but existing in other hospitals):
        t = np.concatenate(
            [
                [species_mapping[f_] for f_ in species_train],
                [species_mapping[f_] for f_ in species_val],
                [species_mapping[f_] for f_ in species_test],
            ]
        )
        for i in set(species_mapping.values()).difference(np.unique(t)):
            species_mapping[i] = nan

        c = 0
        map_to_enum = {nan: nan}
        for k, v in species_mapping.items():
            if v not in map_to_enum:
                map_to_enum[v] = c
                c += 1

        self.species_mapping = {k: map_to_enum[v] for k, v in species_mapping.items()}

        if exclude_nans:
            train_nan = np.array([species_mapping[s] != nan for s in species_train])
            val_nan = np.array([species_mapping[s] != nan for s in species_val])
            test_nan = np.array([species_mapping[s] != nan for s in species_test])

            self.train_indices = self.train_indices[train_nan]
            self.val_indices = self.val_indices[val_nan]
            self.test_indices = self.test_indices[test_nan]

        self.n_species = len(
            np.unique(
                [self.species_mapping[f_] for f_ in f["central"][:][self.train_indices]]
            )
        )

        f.close()

    def setup(self, stage):
        f = h5torch.File(self.path)
        if self.min_spectrum_len is not None:
            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.train_indices]])
            self.train_indices = self.train_indices[lens > self.min_spectrum_len]

            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.val_indices]])
            self.val_indices = self.val_indices[lens > self.min_spectrum_len]

            lens = np.array([len(tt) for tt in f["0/intensity"][:][self.test_indices]])
            self.test_indices = self.test_indices[lens > self.min_spectrum_len]

        if self.in_memory:
            f = f.to_dict()

        self.train = h5torch.Dataset(
            f, subset=self.train_indices, sample_processor=self.processor
        )

        self.val = h5torch.Dataset(
            f, subset=self.val_indices, sample_processor=self.processor
        )

        self.test = h5torch.Dataset(
            f, subset=self.test_indices, sample_processor=self.processor
        )

    def processor(self, f, sample):
        spectrum = SpectrumObject(
            mz=(sample["0/mz"] if "0/mz" in sample else f["unstructured/mz"][:]),
            intensity=sample["0/intensity"],
        )

        if self.preprocessor is not None:
            spectrum = self.preprocessor(spectrum)

        spectrum = {
            "intensity": torch.tensor(spectrum.intensity).float(),
            "mz": torch.tensor(spectrum.mz),
        }

        return (
            spectrum
            | {
                "species": (
                    sample["central"]
                    if self.species_mapping is None
                    else self.species_mapping[sample["central"]]
                )
            }
            | {"loc": sample["0/loc"].astype(str)}
        )


class SpeciesClfDataModule(MaldiDataModule):
    def __init__(
        self,
        path,
        batch_size=512,
        n_workers=4,
        preprocessor=None,
        in_memory=True,
    ):
        super().__init__(batch_size=batch_size, n_workers=n_workers)
        self.path = path
        self.preprocessor = preprocessor
        self.in_memory = in_memory

    def setup(self, stage):
        f = h5torch.File(self.path)

        if self.in_memory:
            f = f.to_dict()

        self.train = MALDITOFDataset(
            f, subset=("unstructured/split", "train"), preprocessor=self.preprocessor
        )

        self.val = MALDITOFDataset(
            f, subset=("unstructured/split", "val"), preprocessor=self.preprocessor
        )

        self.test = MALDITOFDataset(
            f, subset=("unstructured/split", "test"), preprocessor=self.preprocessor
        )

        self.n_species = len(f["unstructured/species_labels"])


def batch_collater(batch):
    batch_collated = {}
    keys = list(batch[0])
    for k in keys:
        v = [b[k] for b in batch]
        if isinstance(v[0], str):
            batch_collated[k] = v
        elif isinstance(v[0], (int, np.int64)):
            batch_collated[k] = torch.tensor(v)
        elif isinstance(v[0], np.ndarray):
            if len({t.shape for t in v}) == 1:
                batch_collated[k] = torch.tensor(np.array(v))
            else:
                batch_collated[k] = pad_sequence(
                    [torch.tensor(t) for t in v], batch_first=True, padding_value=-1
                )
        elif torch.is_tensor(v[0]):
            if len({t.shape for t in v}) == 1:
                batch_collated[k] = torch.stack(v)
            else:
                if v[0].dtype == torch.bool:
                    batch_collated[k] = pad_sequence(
                        v, batch_first=True, padding_value=False
                    )
                else:
                    batch_collated[k] = pad_sequence(
                        v, batch_first=True, padding_value=-1
                    )
    return batch_collated



import os
import argparse
import torch
from sklearn.metrics import roc_auc_score
import numpy as np
from maldi_nn.models import AMRModel, MaldiTransformer
from maldi_nn.utils.data import DRIAMSAMRDataModule
from maldi_nn.spectrum import PeakFilter
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch import Trainer
import ast
from maldi_nn.reproduce.modules import (
    MaldiTransformerNegSampler,
    MaldiTransformerOnlyClf,
    MaldiTransformerMaskMSE
)


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


# eval phase:
def micro_roc_auc(trues, preds, locs, drug_names):
    return roc_auc_score(trues, preds)


def macro_roc_auc(trues, preds, locs, drug_names):
    t = []
    for c in np.unique(drug_names):
        trues_sub = trues[drug_names == c]
        if len(np.unique(trues_sub)) > 1:
            preds_sub = preds[drug_names == c]
            t.append(roc_auc_score(trues_sub, preds_sub))
    return np.mean(t)


def instance_roc_auc(trues, preds, locs, drug_names):
    r, c = np.unique(locs, return_counts=True)
    pos = 0
    total = 0
    for r_ in r[c > 1]:
        trues_sub = trues[locs == r_]
        if len(np.unique(trues[locs == r_])) > 1:
            preds_sub = preds[locs == r_]
            p = preds_sub[trues_sub == 1][:, None] >= preds_sub[trues_sub == 0]
            pos += p.sum()
            total += p.size
    return pos / total


def prec_at_1_neg(trues, preds, locs, drug_names):
    r, c = np.unique(locs, return_counts=True)
    pos = 0
    total = 0
    for r_ in r[c > 1]:
        trues_sub = trues[locs == r_]
        if len(np.unique(trues[locs == r_])) > 1:
            preds_sub = preds[locs == r_]
            pos += trues_sub[np.argmin(preds_sub)]
            total += 1
    return (total - pos) / total

size_to_layer_dims ={
    "mlp" : {
        "S": [256, 128],
        "M": [512, 256, 128],
        "L": [1024, 512, 256, 128],
        "XL": [2048, 1024, 512, 256, 128],
        "Linear": [],
    },
    "trf": {
        "S": [160, 4],
        "M": [184, 6],
        "L": [232, 8],
        "XL": [304, 10],
    }
}

drug_encoder_args = {
    "onehot": {},
    "kernel": {},
    "ecfp": {"bits": 512, "diameter": 4},
    "cnn": {"alphabet": "deepsmiles"},
    "trf": {"alphabet": "deepsmiles"},
    "gru": {"alphabet": "deepsmiles"},
    "img": {"size": (128, 128), "normalize": True, "inverse": True},
}


def main():
    parser = argparse.ArgumentParser(
        description="Training script for dual-branch AMR recommender.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument(
        "logs_path", type=str, metavar="logs_path", help="path to logs."
    )
    parser.add_argument(
        "drug_embedder",
        type=str,
        metavar="drug_embedder",
        choices=["ecfp", "onehot", "gru", "cnn", "trf", "img", "kernel"],
        help="Which drug embedder to use, choices: {%(choices)s}",
    )

    parser.add_argument(
        "spectrum_embedder",
        type=str,
        metavar="spectrum_embedder",
        choices=["trf", "mlp"],
        help="Which spectrum embedder to use, choices: {%(choices)s}",
    )

    parser.add_argument(
        "spectrum_embedder_size",
        type=str,
        metavar="spectrum_embedder_size",
        choices=["S", "M", "L", "XL", "Linear"],
        help="Which size to use for spectrum embedder, choices: {%(choices)s}. Linear is only available for mlp.",
    )
    parser.add_argument("--lr", type=float, default=5e-4, help="Learning rate.")
    parser.add_argument(
        "--logging_file",
        type=str,
        default="res.txt",
        help="Which file to write final performances to.",
    )
    parser.add_argument(
        "--num_workers",
        type=int,
        default=4,
        help="Number of workers in dataloader. Reduce to alleviate CPU.",
    )

    parser.add_argument(
        "--devices",
        type=ast.literal_eval,
        default=1,
        help="devices to use. Input an integer to specify a number of gpus or a list e.g. [1] or [0,1,3] to specify which gpus.",
    )

    parser.add_argument(
        "--trf_n_peaks",
        type=int,
        default=200,
        help="Number of peaks for transformer-peak-based models",
    )

    parser.add_argument(
        "--trf_ckpt_path",
        type=str,
        default="None",
        help="Checkpoint path of malditransformer",
    )
    parser.add_argument(
        "--trf_ckpt_modeltype",
        type=str,
        default="vanilla",
        choices=["vanilla", "negpeaksampler", "intensitymlm", "onlyclf", "onlyshf"],
        help="Maldi Transformer pre-trained modeltype. choices: {%(choices)s}",
    )

    args = parser.parse_args()

    dm = DRIAMSAMRDataModule(
        args.path,
        drug_encoder=args.drug_embedder,
        drug_encoder_args=drug_encoder_args[args.drug_embedder],
        batch_size=128,
        n_workers=args.num_workers,
        preprocessor=(None if args.spectrum_embedder == "mlp" else PeakFilter(max_number=args.trf_n_peaks)),
        min_spectrum_len=None,
        in_memory=True,
    )
    dm.setup(None)

    drug_embedder_kwargs = {
        "onehot": {
            "num_drugs": len(dm.drug_encoder.vocab),
            "dim": 64,
        },
        "kernel": {
            "n_inputs": len(dm.drug_encoder.embeds),
            "n_outputs": 64,
            "layer_dims": [],
            "dropout": 0.2,
        },
        "ecfp": {"n_inputs": 512, "n_outputs": 64, "layer_dims": [], "dropout": 0.2},
        "cnn": {
            "vocab_size": len(dm.drug_encoder.alphabet_dict),
            "dim": 64,
            "depth": 4,
            "kernel_size": 5,
        },
        "trf": {
            "vocab_size": len(dm.drug_encoder.alphabet_dict),
            "dim": 64,
            "depth": 4,
            "n_head": 8,
        },
        "gru": {"vocab_size": len(dm.drug_encoder.alphabet_dict), "hidden_dim": 64},
        "img": {"stem_downsample": 2, "kernel_size": 5, "hidden_dim": 32, "depth": 2},
    }

    if args.spectrum_embedder == "mlp":
        spectrum_kwargs = {
            "n_inputs": 6000,
            "n_outputs": 64,
            "layer_dims": size_to_layer_dims["mlp"][args.spectrum_embedder_size],
            "dropout": 0.2,
        }
    elif args.spectrum_embedder == "trf":
        spectrum_kwargs = {
            "depth": size_to_layer_dims["trf"][args.spectrum_embedder_size][1],
            "dim": size_to_layer_dims["trf"][args.spectrum_embedder_size][0],
            "output_head_dim": 64,
            "reduce": "cls",
            "dropout": 0.2,
        }
        if args.trf_ckpt_path != "None":
            modeltype = MaldiTransformer
            if args.trf_ckpt_modeltype == "intensitymlm":
                modeltype = MaldiTransformerMaskMSE
            elif args.trf_ckpt_modeltype == "negpeaksampler":
                modeltype = MaldiTransformerNegSampler
            elif args.trf_ckpt_modeltype == "onlyclf":
                modeltype = MaldiTransformerOnlyClf
            

            malditransformer = modeltype.load_from_checkpoint(args.trf_ckpt_path)
            spectrum_kwargs["depth"] = malditransformer.hparams.depth
            spectrum_kwargs["dim"] = malditransformer.hparams.dim

    model = AMRModel(
        spectrum_embedder=args.spectrum_embedder,
        drug_embedder=args.drug_embedder,
        spectrum_kwargs=spectrum_kwargs,
        drug_kwargs=drug_embedder_kwargs[args.drug_embedder],
        lr=args.lr,
        weight_decay=0,
        lr_decay_factor=1,
        warmup_steps=250,
        scaled_dot_product=True,
    )

    if (args.trf_ckpt_path != "None") and (args.spectrum_embedder == "trf"):
        pretrained_dict = malditransformer.transformer.state_dict()
        model_state_dict = model.spectrum_embedder.state_dict()
        pretrained_dict = {
            k: v for k, v in pretrained_dict.items() if not k.startswith("output_head")
        }
        model_state_dict.update(pretrained_dict)
        model.spectrum_embedder.load_state_dict(model_state_dict)


    val_ckpt = ModelCheckpoint(monitor="val_micro_auc", mode="max")
    callbacks = [
        val_ckpt,
        EarlyStopping(monitor="val_micro_auc", patience=10, mode="max"),
    ]
    logger = TensorBoardLogger(
        args.logs_path,
        name="amr%s_%s_%s_%s" % (args.spectrum_embedder, args.drug_embedder, args.spectrum_embedder_size, args.lr),
    )

    trainer = Trainer(
        accelerator="gpu",
        devices=args.devices,
        strategy="auto",
        val_check_interval=0.1,
        max_epochs=50,
        callbacks=callbacks,
        logger=logger,
    )

    trainer.fit(model, dm.train_dataloader(), dm.val_dataloader())

    # validation
    p = trainer.predict(
        model, dataloaders=dm.val_dataloader(), ckpt_path=val_ckpt.best_model_path
    )

    trues = torch.cat([pp[0][:, 0] for pp in p]).numpy()
    preds = torch.cat([pp[0][:, 1] for pp in p]).numpy()
    locs = np.concatenate([pp[1] for pp in p])
    drug_names = np.concatenate([pp[2] for pp in p])

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\tval\t%.5f\t%.5f\t%.5f\t%.5f\n"
            % (
                val_ckpt.best_model_path,
                micro_roc_auc(trues, preds, locs, drug_names),
                macro_roc_auc(trues, preds, locs, drug_names),
                instance_roc_auc(trues, preds, locs, drug_names),
                prec_at_1_neg(trues, preds, locs, drug_names),
            )
        )

    # test
    p = trainer.predict(
        model, dataloaders=dm.test_dataloader(), ckpt_path=val_ckpt.best_model_path
    )

    trues = torch.cat([pp[0][:, 0] for pp in p]).numpy()
    preds = torch.cat([pp[0][:, 1] for pp in p]).numpy()
    locs = np.concatenate([pp[1] for pp in p])
    drug_names = np.concatenate([pp[2] for pp in p])

    with open(os.path.join(args.logs_path, args.logging_file), "a") as f:
        f.write(
            "%s\ttest\t%.5f\t%.5f\t%.5f\t%.5f\n"
            % (
                val_ckpt.best_model_path,
                micro_roc_auc(trues, preds, locs, drug_names),
                macro_roc_auc(trues, preds, locs, drug_names),
                instance_roc_auc(trues, preds, locs, drug_names),
                prec_at_1_neg(trues, preds, locs, drug_names),
            )
        )


if __name__ == "__main__":
    main()



from rdkit import Chem
from rdkit.Chem.Draw import MolToImage
from rdkit.Chem import AllChem
import numpy as np
import deepsmiles
import selfies
import re
import torch
import maldi_nn.nn as maldinn
import torch.nn as nn
from bio_attention.attention import GLU, TransformerEncoder, Aggregator
from bio_attention.embed import DiscreteEmbedding


class SmilesEncoder:
    def __init__(self):
        self.vocab = {}
        self.embeds = []
        self.alphabet_dict = {}

    def __call__(self, smiles_string):
        return (
            self.vocab[smiles_string]
            if smiles_string in self.vocab
            else self.embed(smiles_string)
        )

    def embed(self, smiles_string):
        return NotImplementedError

    def create_vocab(self, smiles_list):
        return {smiles: self.embed(smiles) for smiles in smiles_list}


class SmilesToIndex(SmilesEncoder):
    def __init__(self, smiles_list):
        super().__init__()
        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        raise NotImplementedError("Can't embed new SMILES strings.")

    def create_vocab(self, smiles_list):
        return {smiles: ix for ix, smiles in enumerate(smiles_list)}


class SmilesToECFP(SmilesEncoder):
    def __init__(self, smiles_list, bits=1024, diameter=4):
        super().__init__()

        self.bits = bits
        self.rad = int(diameter / 2)

        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        molecule = Chem.MolFromSmiles(smiles_string)
        return np.array(
            AllChem.GetMorganFingerprintAsBitVect(molecule, self.rad, nBits=self.bits)
        ).astype(np.float32)


class SmilesToAlphabet(SmilesEncoder):
    def __init__(self, smiles_list):
        super().__init__()

        alphabet = set([c for smiles in smiles_list for c in smiles])
        alphabet = sorted(alphabet)
        self.alphabet_dict = {a: i for i, a in enumerate(alphabet)}
        self.alphabet_dict["[UNK]"] = len(self.alphabet_dict)
        self.onehot_selector = np.eye(len(self.alphabet_dict))

        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        smiles_ints = self.string2onehot(smiles_string)
        return self.onehot_selector[smiles_ints].astype(np.float32)

    def string2onehot(self, string):
        return np.array(
            [
                self.alphabet_dict[c]
                if c in self.alphabet_dict
                else self.alphabet_dict["[UNK]"]
                for c in string
            ]
        )


class SmilesToDEEPSmilesAlphabet(SmilesEncoder):
    def __init__(self, smiles_list):
        super().__init__()
        self.converter = deepsmiles.Converter(rings=True, branches=True)

        alphabet = set(
            [c for smiles in smiles_list for c in self.smiles_to_deepsmiles(smiles)]
        )
        alphabet = sorted(alphabet)
        self.alphabet_dict = {a: i for i, a in enumerate(alphabet)}
        self.alphabet_dict["[UNK]"] = len(self.alphabet_dict)
        self.onehot_selector = np.eye(len(self.alphabet_dict))

        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        smiles_ints = self.string2onehot(smiles_string)
        return self.onehot_selector[smiles_ints].astype(np.float32)

    def string2onehot(self, string):
        return np.array(
            [
                self.alphabet_dict[c]
                if c in self.alphabet_dict
                else self.alphabet_dict["[UNK]"]
                for c in self.smiles_to_deepsmiles(string)
            ]
        )

    def smiles_to_deepsmiles(self, smiles):
        return self.converter.encode(smiles)


class SmilesToSelfiesAlphabet(SmilesEncoder):
    def __init__(self, smiles_list):
        super().__init__()

        alphabet = selfies.get_alphabet_from_selfies(
            [selfies.encoder(smiles) for smiles in smiles_list]
        )
        alphabet = sorted(alphabet)
        self.alphabet_dict = {a: i for i, a in enumerate(alphabet)}
        self.alphabet_dict["."] = len(self.alphabet_dict)
        self.alphabet_dict["[UNK]"] = len(self.alphabet_dict)
        self.onehot_selector = np.eye(len(self.alphabet_dict))

        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        smiles_ints = self.string2onehot(smiles_string)
        return self.onehot_selector[smiles_ints].astype(np.float32)

    def string2onehot(self, string):
        return np.array(
            [
                self.alphabet_dict[c]
                if c in self.alphabet_dict
                else self.alphabet_dict["[UNK]"]
                for c in selfies.split_selfies(self.smiles_to_selfies(string))
            ]
        )

    def smiles_to_selfies(self, smiles):
        return selfies.encoder(smiles)


class SmilesToImage(SmilesEncoder):
    def __init__(self, smiles_list, size=(150, 150), normalize=True, inverse=False):
        super().__init__()
        self.size = size
        self.normalize = normalize
        self.inverse = inverse
        self.vocab = self.create_vocab(smiles_list)

    def embed(self, smiles_string):
        mol = Chem.MolFromSmiles(smiles_string)
        embed = np.array(MolToImage(mol, size=self.size))
        if self.inverse:
            embed = 255 - embed
        if self.normalize:
            embed = embed / 255

        return embed.transpose(2, 0, 1).astype(np.float32)


class SmilesToLINGOKernel(SmilesEncoder):
    def __init__(self, smiles_list):
        super().__init__()
        embedded = [self.get_n_fourmers(smiles) for smiles in smiles_list]

        kernel = np.zeros((len(embedded), len(embedded)))
        for i in range(len(embedded)):
            for j in range(i, len(embedded)):
                sim = self.LINGOsim(embedded[i], embedded[j])
                kernel[i, j] = sim
                kernel[j, i] = sim

        self.vocab = {smiles: kernel[ix, :] for ix, smiles in enumerate(smiles_list)}
        self.embeds = embedded

    def embed(self, smiles_string):
        embedded = self.get_n_fourmers(smiles_string)
        return np.array([self.LINGOsim(embedded, n) for n in self.embeds])

    def get_n_fourmers(self, smiles):
        smiles = re.sub("\d", "o", smiles)
        fourmers = [smiles[s : s + 4] for s in range(len(smiles) - 3)]
        return {s: c for s, c in zip(*np.unique(fourmers, return_counts=True))}

    def LINGOsim(self, s1, s2):
        l_ = set.union(*[set(s1.keys()), set(s2.keys())])
        return sum(
            [
                1
                - abs(self.take(s1, l) - self.take(s2, l))
                / abs(self.take(s1, l) + self.take(s2, l))
                for l in l_
            ]
        ) / len(l_)

    @staticmethod
    def take(dict_, elem):
        return dict_[elem] if elem in dict_ else 0


class DrugOneHotEmbedding(nn.Module):
    def __init__(self, num_drugs, dim):
        super().__init__()
        self.embed = nn.Embedding(num_drugs, dim)

    def forward(self, drug):
        return self.embed(drug.long())


class DrugMLP(maldinn.MLP):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, drug):
        return self.net(drug)


class DrugGRU(nn.Module):
    def __init__(self, vocab_size, hidden_dim=256, bidirectional=True, out_head=64):
        super().__init__()
        self.GRU = nn.GRU(
            vocab_size,
            hidden_dim,
            num_layers=1,
            bidirectional=bidirectional,
            batch_first=True,
        )
        self.output_head = nn.Linear(hidden_dim, out_head)

    def forward(self, drug):
        mask = (drug != -1).all(-1)
        x = nn.utils.rnn.pack_padded_sequence(
            drug, mask.sum(1).tolist(), batch_first=True, enforce_sorted=False
        )
        _, out = self.GRU(x)

        return self.output_head(out.mean(0))


class MaskedConv1dWrapper(nn.Module):
    def __init__(self, ConvModule):
        """
        Works only for Conv1ds that have the same input shape as output shape.
        """
        assert isinstance(ConvModule, nn.Conv1d)
        super().__init__()
        self.conv = ConvModule

    def forward(self, x, mask=None):
        """
        x = (B, C, L)
        mask = (B, L)
        """
        if mask is not None:
            return self.conv(x * mask.unsqueeze(1))
        else:
            return self.conv(x)


class Conv1DBlock(nn.Module):
    def __init__(
        self,
        dim,
        kernel_size,
        depthwise=False,
        glu_ff=True,
        dropout=0.2,
        activation="gelu",
    ):
        super().__init__()
        if activation.lower() == "relu":
            act = nn.ReLU()
        elif activation.lower() == "gelu":
            act = nn.GELU()
        elif activation.lower() == "swish":
            act = nn.SiLU()

        self.conv = MaskedConv1dWrapper(
            nn.Conv1d(
                dim,
                dim,
                kernel_size=kernel_size,
                padding=kernel_size // 2,
                groups=(dim if depthwise else 1),
            )
        )

        self.norm = nn.Sequential(
            maldinn.Permute(0, 2, 1),
            nn.LayerNorm(dim),
            maldinn.Permute(0, 2, 1),
        )

        self.prenorm = nn.Sequential(
            maldinn.Permute(0, 2, 1),
            nn.LayerNorm(dim),
            maldinn.Permute(0, 2, 1),
        )

        project_in = (
            nn.Sequential(nn.Linear(dim, 4 * dim), act)
            if not glu_ff
            else GLU(dim, 4 * dim, act)
        )

        self.pointwise_net = nn.Sequential(
            maldinn.Permute(0, 2, 1),
            project_in,
            nn.Dropout(dropout),
            nn.Linear(4 * dim, dim),
            maldinn.Permute(0, 2, 1),
        )

    def forward(self, x, mask=None):
        z = self.conv(self.prenorm(x), mask=mask) + x
        z = self.pointwise_net(self.norm(z)) + z
        return z


class DrugCNN(nn.Module):
    def __init__(
        self,
        vocab_size,
        dim=64,
        depth=4,
        kernel_size=5,
        dropout=0.2,
        out_head=64,
    ):
        super().__init__()

        self.stem = nn.Sequential(nn.Linear(vocab_size, dim), maldinn.Permute(0, 2, 1))

        body = []
        for _ in range(depth):
            body.append(
                Conv1DBlock(
                    dim,
                    kernel_size=kernel_size,
                    dropout=dropout,
                    depthwise=False,
                    glu_ff=True,
                    activation="gelu",
                )
            )
        self.body = nn.ModuleList(body)
        self.output_head = nn.Linear(dim, out_head)

    def forward(self, drug):
        mask = (drug != -1).all(-1)
        z = self.stem(drug)
        for layer in self.body:
            z = layer(z, mask=mask)

        return self.output_head(z.max(2).values)


class DrugTransformer(nn.Module):
    def __init__(
        self,
        vocab_size,
        dim=128,
        depth=6,
        n_head=8,
        dropout=0.2,
        out_head=64,
    ):
        super().__init__()

        self.embed = DiscreteEmbedding(vocab_size, dim, cls=True)
        self.body = TransformerEncoder(
            depth=depth,
            dim=dim,
            nh=n_head,
            attention_args={"dropout": dropout},
            plugintype="sinusoidal",
            plugin_args={"dim": dim},
            only_apply_plugin_at_first=True,
            dropout=dropout,
            glu_ff=True,
            activation="gelu",
        )

        self.agg = Aggregator(method="cls")
        self.output_head = nn.Linear(dim, out_head)

    def forward(self, drug):
        mask = (drug != -1).all(-1)
        drug = torch.argmax(drug, dim=2).to(self.embed.embedder.weight.dtype)
        drug[~mask] = torch.nan

        x = self.embed(drug)

        z = self.body(x, mask=mask)

        return self.output_head(self.agg(z))


class Conv2DBlock(nn.Module):
    def __init__(
        self,
        dim,
        kernel_size=5,
        dropout=0.2,
        depthwise=False,
        glu_ff=True,
        activation="gelu",
    ):
        super().__init__()

        if activation.lower() == "relu":
            act = nn.ReLU()
        elif activation.lower() == "gelu":
            act = nn.GELU()
        elif activation.lower() == "swish":
            act = nn.SiLU()

        self.conv = nn.Conv2d(
            dim,
            dim,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
            groups=(dim if depthwise else 1),
        )

        self.norm = nn.Sequential(
            maldinn.Permute(0, 2, 3, 1),
            nn.LayerNorm(dim),
            maldinn.Permute(0, 3, 1, 2),
        )

        project_in = (
            nn.Sequential(nn.Linear(dim, 4 * dim), act)
            if not glu_ff
            else GLU(dim, 4 * dim, act)
        )

        self.pointwise_net = nn.Sequential(
            maldinn.Permute(0, 2, 3, 1),
            project_in,
            nn.Dropout(dropout),
            nn.Linear(4 * dim, dim),
            maldinn.Permute(0, 3, 1, 2),
        )

        self.prenorm = nn.Sequential(
            maldinn.Permute(0, 2, 3, 1),
            nn.LayerNorm(dim),
            maldinn.Permute(0, 3, 1, 2),
        )

    def forward(self, x):
        z = self.conv(self.prenorm(x)) + x
        z = self.pointwise_net(self.norm(z)) + z
        return z


class DrugImageCNN(nn.Module):
    def __init__(
        self,
        stem_downsample=2,
        kernel_size=5,
        hidden_dim=32,
        depth=2,
        dropout=0.2,
        out_head=64,
    ):
        super().__init__()

        body = [nn.Conv2d(3, hidden_dim, stem_downsample, stride=stem_downsample)]

        for _ in range(depth):
            body.append(
                Conv2DBlock(
                    hidden_dim,
                    kernel_size=kernel_size,
                    dropout=dropout,
                    glu_ff=True,
                    activation="gelu",
                    depthwise=False,
                )
            )

        self.body = nn.Sequential(*body)

        self.output_head = nn.Linear(hidden_dim, out_head)

    def forward(self, drug):
        z = self.body(drug)
        return self.output_head(z.max(-1).values.max(-1).values)



from maldi_nn.utils.data import DRIAMSSpectrumDataModule
import torch
from maldi_nn.spectrum import *
import h5torch
import sys
import argparse

def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass

def main():
    parser = argparse.ArgumentParser(
        description="Estimate distribution over negative peaks, and add it to the h5torch file. Necessary to run before training Maldiformer using sampled negative peaks.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument("path", type=str, metavar="path", help="path to h5torch file.")
    parser.add_argument(
        "--n_peaks",
        type=int,
        default=200,
        help="Number of peaks",
    )
    args = parser.parse_args()
    
    dm = DRIAMSSpectrumDataModule(
        args.path,
        batch_size=128,
        n_workers=8,
        preprocessor=SequentialPreprocessor(
            PeakFilter(args.n_peaks), Binner(start=2000, stop=20000, step=1)
        ),
        min_spectrum_len=128,
        exclude_nans=False,
        in_memory=False,
    )
    f = h5torch.File(dm.path)
    if dm.min_spectrum_len is not None:
        lens = np.array([len(tt) for tt in f["0/intensity"][:][dm.train_indices]])
        dm.train_indices = dm.train_indices[lens > dm.min_spectrum_len]

        lens = np.array([len(tt) for tt in f["0/intensity"][:][dm.val_indices]])
        dm.val_indices = dm.val_indices[lens > dm.min_spectrum_len]

        lens = np.array([len(tt) for tt in f["0/intensity"][:][dm.test_indices]])
        dm.test_indices = dm.test_indices[lens > dm.min_spectrum_len]


    f2 = f.to_dict()
    f.close()

    dm.train = h5torch.Dataset(f2, subset=dm.train_indices, sample_processor=dm.processor)

    p_x = torch.zeros(dm.train[0]["intensity"].shape)
    p_y_x = [torch.tensor([]) for _ in range(len(p_x))]
    dl = torch.utils.data.DataLoader(
        dm.train,
        num_workers=8,
        batch_size=128,
        shuffle=True,
        pin_memory=False,
    )

    len_ = len(dl)
    print(len_)
    for ix, s in enumerate(iter(dl)):
        p_x += (s["intensity"] != 0).sum(0)

        a = torch.unique(torch.nonzero(s["intensity"])[:, 1])
        for aa in a:
            sub = s["intensity"][:, aa]
            p_y_x[aa] = torch.cat([p_y_x[aa], sub[sub != 0]])
        if (ix % 25) == 0:
            print(ix)

    p_x = (p_x / p_x.sum()).numpy()

    p_y_x_q = []
    for p in p_y_x:
        if len(p) > 0:
            p_y_x_q.append(torch.quantile(p, torch.linspace(0, 1, 101)))
        else:
            p_y_x_q.append(torch.zeros(101))


    p_y_x = torch.stack(p_y_x_q).numpy()

    f = h5torch.File(args.path, "a")
    f.create_dataset("unstructured/p_x_200", data=p_x)
    f.create_dataset("unstructured/p_y_x_200", data=p_y_x)
    f.close()


if __name__ == "__main__":
    main()
